"""
Functions for plotting datasets nicely.
"""
# TODO: custom xtick labels                                                   #
# TODO: annotations, arbitrary text                                           #
# TODO: docs                                                                  #

import functools

import numpy as np

from ..manage import auto_xyz_ds
from ..utils import autocorrect_kwargs
from .color import xyz_colormaps
from .core import (
    PLOTTER_DEFAULTS,
    AbstractHeatMap,
    AbstractHistogram,
    AbstractLinePlot,
    AbstractScatter,
    Plotter,
    calc_row_col_datasets,
    intercept_call_arg,
    prettify,
)

# ----------------- Main lineplot interface for matplotlib ------------------ #


class PlotterMatplotlib(Plotter):
    """ """

    def __init__(self, ds, x, y, z=None, y_err=None, x_err=None, **kwargs):
        super().__init__(
            ds,
            x,
            y,
            z=z,
            y_err=y_err,
            x_err=x_err,
            **kwargs,
            backend="MATPLOTLIB",
        )

    def prepare_axes(self):
        """ """
        import matplotlib as mpl

        if self.math_serif:
            mpl.rcParams["mathtext.fontset"] = "cm"
            mpl.rcParams["mathtext.rm"] = "serif"
        mpl.rcParams["font.family"] = self.font
        mpl.rcParams["font.weight"] = self.font_weight
        import matplotlib.pyplot as plt

        if self.axes_rloc is not None:
            if self.axes_loc is not None:
                raise ValueError(
                    "Cannot specify absolute and relative "
                    "location of axes at the same time."
                )
            if self.add_to_fig is None:
                raise ValueError(
                    "Can only specify relative axes position "
                    "when adding to a figure, i.e. when "
                    "add_to_fig != None"
                )

        if self.axes_rloc is not None:
            self._axes_loc = self._cax_rel2abs_rect(
                self.axes_rloc, self.add_to_fig.get_axes()[-1]
            )
        else:
            self._axes_loc = self.axes_loc

        # Add a new set of axes to an existing plot
        if self.add_to_fig is not None and self.subplot is None:
            self._fig = self.add_to_fig
            self._axes = self._fig.add_axes(
                (0.4, 0.6, 0.30, 0.25)
                if self._axes_loc is None
                else self._axes_loc
            )

        # Add lines to an existing set of axes
        elif self.add_to_axes is not None:
            self._fig = self.add_to_axes
            self._axes = self._fig.get_axes()[-1]

        # Add lines to existing axes but only sharing the x-axis
        elif self.add_to_xaxes is not None:
            self._fig = self.add_to_xaxes
            self._axes = self._fig.get_axes()[-1].twinx()

        elif self.subplot is not None:
            # Add new axes as subplot to existing subplot
            if self.add_to_fig is not None:
                self._fig = self.add_to_fig

            # New figure but add as subplot
            else:
                self._fig = plt.figure(
                    self.fignum, figsize=self.figsize, dpi=100
                )
            self._axes = self._fig.add_subplot(self.subplot)

        # Make new figure and axes
        else:
            self._fig = plt.figure(self.fignum, figsize=self.figsize, dpi=100)
            self._axes = self._fig.add_axes(
                (0.15, 0.15, 0.8, 0.75)
                if self._axes_loc is None
                else self._axes_loc
            )
        self._axes.set_title(
            "" if self.title is None else self.title,
            fontsize=self.fontsize_title,
            pad=self.title_pad,
        )

    def set_axes_labels(self):
        if self._xtitle:
            self._axes.set_xlabel(self._xtitle, fontsize=self.fontsize_xtitle)
            self._axes.xaxis.labelpad = self.xtitle_pad
        if self._ytitle:
            self._axes.set_ylabel(self._ytitle, fontsize=self.fontsize_ytitle)
            self._axes.yaxis.labelpad = self.ytitle_pad
            if self.ytitle_right:
                self._axes.yaxis.set_label_position("right")

    def set_axes_scale(self):
        """ """
        self._axes.set_xscale("log" if self.xlog else "linear")
        self._axes.set_yscale("log" if self.ylog else "linear")

    def set_axes_range(self):
        """ """
        if self._xlims:
            self._axes.set_xlim(self._xlims)
        if self._ylims:
            self._axes.set_ylim(self._ylims)

    def set_spans(self):
        """ """
        if self.vlines is not None:
            for x in self.vlines:
                self._axes.axvline(
                    x,
                    lw=self.span_width,
                    color=self.span_color,
                    linestyle=self.span_style,
                )
        if self.hlines is not None:
            for y in self.hlines:
                self._axes.axhline(
                    y,
                    lw=self.span_width,
                    color=self.span_color,
                    linestyle=self.span_style,
                )

    def set_gridlines(self):
        """ """
        for axis in ("top", "bottom", "left", "right"):
            self._axes.spines[axis].set_linewidth(1.0)

        if self.gridlines:
            # matplotlib has coarser gridine style than bokeh
            self._gridline_style = [x / 2 for x in self.gridline_style]
            self._axes.set_axisbelow(True)  # ensures gridlines below all
            self._axes.grid(
                True,
                color="0.9",
                which="major",
                linestyle=(0, self._gridline_style),
            )
            self._axes.grid(
                True,
                color="0.95",
                which="minor",
                linestyle=(0, self._gridline_style),
            )

    def set_tick_marks(self):
        """ """
        import matplotlib as mpl

        if self.xticks is not None:
            self._axes.set_xticks(self.xticks, minor=False)
            self._axes.get_xaxis().set_major_formatter(
                mpl.ticker.ScalarFormatter()
            )
        if self.yticks is not None:
            self._axes.set_yticks(self.yticks, minor=False)
            self._axes.get_yaxis().set_major_formatter(
                mpl.ticker.ScalarFormatter()
            )

        if self.xtick_labels is not None:
            self._axes.set_xticklabels(self.xtick_labels)

        if self.xticklabels_hide:
            (
                self._axes.get_xaxis().set_major_formatter(
                    mpl.ticker.NullFormatter()
                )
            )
        if self.yticklabels_hide:
            (
                self._axes.get_yaxis().set_major_formatter(
                    mpl.ticker.NullFormatter()
                )
            )

        self._axes.tick_params(
            labelsize=self.fontsize_ticks,
            direction="out",
            bottom="bottom" in self.ticks_where,
            top="top" in self.ticks_where,
            left="left" in self.ticks_where,
            right="right" in self.ticks_where,
        )

        if self.yticklabels_right or (
            self.yticklabels_right is None and self.ytitle_right is True
        ):
            self._axes.yaxis.tick_right()

    def _cax_rel2abs_rect(self, rel_rect, cax=None):
        """Turn a relative axes specification into a absolute one."""
        if cax is None:
            cax = self._axes
        bbox = cax.get_position()
        l, b, w, h = bbox.x0, bbox.y0, bbox.width, bbox.height
        cl = l + w * rel_rect[0]
        cb = b + h * rel_rect[1]
        try:
            cw = w * rel_rect[2]
            ch = h * rel_rect[3]
        except IndexError:
            return cl, cb
        return cl, cb, cw, ch

    def plot_legend(self, grid=False, labels_handles=None):
        """Add a legend"""
        if self._use_legend:
            if labels_handles:
                labels, handles = zip(*labels_handles.items())
            else:
                handles, labels = self._legend_handles, self._legend_labels

            if self.legend_reverse:
                handles, labels = handles[::-1], labels[::-1]

            # Limit minimum size of markers that appear in legend
            should_auto_scale_legend_markers = (
                (self.legend_marker_scale is None)  # not already set
                and hasattr(self, "_marker_size")  # is a valid parameter
                and self._marker_size < 3  # and is small
            )
            if should_auto_scale_legend_markers:
                self.legend_marker_scale = 3 / self._marker_size

            opts = {
                "title": (self.z_coo if self.ztitle is None else self.ztitle),
                "loc": self.legend_loc,
                "fontsize": self.fontsize_zlabels,
                "frameon": self.legend_frame,
                "framealpha": self.legend_frame_alpha,
                "numpoints": 1,
                "scatterpoints": 1,
                "handlelength": self.legend_handlelength,
                "markerscale": self.legend_marker_scale,
                "labelspacing": self.legend_label_spacing,
                "columnspacing": self.legend_column_spacing,
                "bbox_to_anchor": self.legend_bbox,
                "ncol": self.legend_ncol,
            }

            if grid:
                bb = opts["bbox_to_anchor"]
                if bb is None:
                    opts["bbox_to_anchor"] = (1, 0.5, 0, 0)
                    opts["loc"] = "center left"
                else:
                    loc = opts["loc"]
                    # will get warning for 'best'
                    opts["loc"] = "center" if loc in ("best", 0) else loc
                lgnd = self._fig.legend(handles, labels, **opts)
            else:
                lgnd = self._axes.legend(handles, labels, **opts)

            lgnd.get_title().set_fontsize(self.fontsize_ztitle)

            if self.legend_marker_alpha is not None:
                for legendline in lgnd.legendHandles:
                    legendline.set_alpha(1.0)

    def set_mappable(self):
        """Mappale object for colorbars."""
        from matplotlib.cm import ScalarMappable

        self.mappable = ScalarMappable(cmap=self.cmap, norm=self._color_norm)
        self.mappable.set_array([])

    def plot_colorbar(self, grid=False):
        """Add a colorbar to the data."""

        if self._use_colorbar:
            # Whether the colorbar should clip at either end
            extendmin = (self.vmin is not None) and (self.vmin > self._zmin)
            extendmax = (self.vmax is not None) and (self.vmax < self._zmax)
            extend = (
                "both"
                if extendmin and extendmax
                else "min"
                if extendmin
                else "max"
                if extendmax
                else "neither"
            )

            opts = {"extend": extend, "ticks": self.zticks}

            if self.colorbar_relative_position:
                opts["cax"] = self._fig.add_axes(
                    self._cax_rel2abs_rect(self.colorbar_relative_position)
                )

            if grid:
                opts["ax"] = self._fig.axes
                opts["anchor"] = (0.5, 0.5)

            self._cbar = self._fig.colorbar(
                self.mappable, **opts, **self.colorbar_opts
            )

            self._cbar.ax.tick_params(labelsize=self.fontsize_zlabels)

            self._cbar.ax.set_title(
                self._ctitle,
                fontsize=self.fontsize_ztitle,
                color=self.colorbar_color if self.colorbar_color else None,
            )

            if self.colorbar_color:
                self._cbar.ax.yaxis.set_tick_params(
                    color=self.colorbar_color, labelcolor=self.colorbar_color
                )
                self._cbar.outline.set_edgecolor(self.colorbar_color)

    def set_panel_label(self):
        if self.panel_label is not None:
            self._axes.text(
                *self.panel_label_loc,
                self.panel_label,
                transform=self._axes.transAxes,
                fontsize=self.fontsize_panel_label,
                color=self.panel_label_color,
                ha="left",
                va="top",
            )

    def show(self):
        import matplotlib.pyplot as plt

        if self.return_fig:
            plt.close(self._fig)
            return self._fig

    def prepare_plot(self):
        """Do all the things that every plot has."""
        self.prepare_axes()
        self.set_axes_labels()
        self.set_axes_scale()
        self.set_axes_range()
        self.set_spans()
        self.set_gridlines()
        self.set_tick_marks()


# --------------------------------------------------------------------------- #


def mpl_multi_plot(fn):
    """Decorate a plotting function to plot a grid of values."""

    @functools.wraps(fn)
    def multi_plotter(
        ds,
        *args,
        row=None,
        col=None,
        hspace=None,
        wspace=None,
        tight_layout=True,
        coltitle=None,
        rowtitle=None,
        **kwargs,
    ):
        if (row is None) and (col is None):
            return fn(ds, *args, **kwargs)

        import matplotlib.pyplot as plt
        from matplotlib.gridspec import GridSpec

        # Set some global parameters
        p = fn(ds, *args, **kwargs, call=False)
        p.prepare_data_multi_grid()

        kwargs["vmin"] = kwargs.pop("vmin", p.vmin)
        kwargs["vmax"] = kwargs.pop("vmax", p.vmax)

        coltitle = col if coltitle is None else coltitle
        rowtitle = row if rowtitle is None else rowtitle

        # split the dataset into its respective rows and columns
        ds_r_c, nrows, ncols = calc_row_col_datasets(ds, row=row, col=col)

        figsize = kwargs.pop("figsize", (3 * ncols, 3 * nrows))
        return_fig = kwargs.pop("return_fig", PLOTTER_DEFAULTS["return_fig"])

        # generate a figure for all the plots to use
        p._fig = plt.figure(
            figsize=figsize, dpi=100, constrained_layout=tight_layout
        )
        p._fig.set_constrained_layout_pads(hspace=hspace, wspace=wspace)
        # and a gridspec to position them
        gs = GridSpec(
            nrows=nrows,
            ncols=ncols,
            figure=p._fig,
            hspace=hspace,
            wspace=wspace,
        )

        # want to collect all entries for legend
        labels_handles = {}

        # range through rows and do subplots
        for i, ds_r in enumerate(ds_r_c):
            skws = {"legend": False, "colorbar": False}

            # if not last row
            if i != nrows - 1:
                skws["xticklabels_hide"] = True
                skws["xtitle"] = ""

            # range through columns
            for j, sub_ds in enumerate(ds_r):
                if hspace == 0 and wspace == 0:
                    ticks_where = []
                    if j == 0:
                        ticks_where.append("left")
                    if i == 0:
                        ticks_where.append("top")
                    if j == ncols - 1:
                        ticks_where.append("right")
                    if i == nrows - 1:
                        ticks_where.append("bottom")
                    skws["ticks_where"] = ticks_where

                # if not first column
                if j != 0:
                    skws["yticklabels_hide"] = True
                    skws["ytitle"] = ""

                # label each column
                if (i == 0) and (col is not None):
                    col_val = prettify(ds[col].values[j])
                    skws["title"] = "{} = {}".format(coltitle, col_val)
                    fx = "fontsize_xtitle"
                    skws["fontsize_title"] = kwargs.get(
                        fx, PLOTTER_DEFAULTS[fx]
                    )

                # label each row
                if (j == ncols - 1) and (row is not None):
                    # XXX: if number of cols==1 this hide yaxis - want both
                    row_val = prettify(ds[row].values[i])
                    skws["ytitle_right"] = True
                    skws["ytitle"] = "{} = {}".format(rowtitle, row_val)

                sP = fn(
                    sub_ds,
                    *args,
                    add_to_fig=p._fig,
                    call="both",
                    subplot=gs[i, j],
                    **{**kwargs, **skws},
                )

                try:
                    labels_handles.update(
                        dict(zip(sP._legend_labels, sP._legend_handles))
                    )
                except AttributeError:
                    pass

        # make sure all have the same plot ranges
        xmins, xmaxs = zip(*(gax.get_xlim() for gax in p._fig.axes))
        ymins, ymaxs = zip(*(gax.get_ylim() for gax in p._fig.axes))
        xmin, xmax = min(xmins), max(xmaxs)
        ymin, ymax = min(ymins), max(ymaxs)
        for gax in p._fig.axes:
            gax.set_xlim(xmin, xmax)
            gax.set_ylim(ymin, ymax)

        # add global legend or colorbar
        p.plot_legend(grid=True, labels_handles=labels_handles)
        p.plot_colorbar(grid=True)

        if return_fig:
            plt.close(p._fig)
            return p._fig

    return multi_plotter


# --------------------------------------------------------------------------- #


class LinePlot(PlotterMatplotlib, AbstractLinePlot):
    """ """

    def __init__(self, ds, x, y, z=None, *, y_err=None, x_err=None, **kwargs):
        super().__init__(ds, x, y, z=z, y_err=y_err, x_err=x_err, **kwargs)

    def plot_lines(self):
        """ """
        for data in self._gen_xy():
            col = next(self._cols)

            line_opts = {
                "c": col,
                "lw": next(self._lws),
                "marker": next(self._mrkrs),
                "markersize": self._marker_size,
                "markeredgecolor": col[:3] + (self.marker_alpha * col[3],),
                "markerfacecolor": col[:3] + (self.marker_alpha * col[3] / 2,),
                "label": next(self._zlbls),
                "zorder": next(self._zordrs),
                "linestyle": next(self._lines),
                "rasterized": self.rasterize,
            }

            if ("ye" in data) or ("xe" in data):
                self._axes.errorbar(
                    data["x"],
                    data["y"],
                    yerr=data.get("ye", None),
                    xerr=data.get("xe", None),
                    ecolor=col,
                    capsize=self.errorbar_capsize,
                    capthick=self.errorbar_capthick,
                    elinewidth=self.errorbar_linewidth,
                    **line_opts,
                )
            else:
                # add line to axes, with options cycled through
                self._axes.plot(data["x"], data["y"], **line_opts)

            (
                self._legend_handles,
                self._legend_labels,
            ) = self._axes.get_legend_handles_labels()

    def __call__(self):
        self.prepare_data_single()
        # matplotlib preparation
        self.prepare_plot()
        self.plot_lines()
        self.plot_legend()
        self.plot_colorbar()
        self.set_panel_label()
        return self.show()


@mpl_multi_plot
@intercept_call_arg
def lineplot(ds, x, y, z=None, y_err=None, x_err=None, **plot_opts):
    """From  ``ds`` plot lines of ``y`` as a function of ``x``, optionally for
    varying ``z``.

    Parameters
    ----------
    ds : xarray.Dataset
        Dataset to plot from.
    x : str
        Dimension to plot along the x-axis.
    y : str or tuple[str]
        Variable(s) to plot along the y-axis. If tuple, plot each of the
        variables - instead of ``z``.
    z : str, optional
        Dimension to plot into the page.
    y_err : str, optional
        Variable to plot as y-error.
    x_err : str, optional
        Variable to plot as x-error.
    row : str, optional
        Dimension to vary over as a function of rows.
    col : str, optional
        Dimension to vary over as a function of columns.
    plot_opts
        See ``xyzpy.plot.core.PLOTTER_DEFAULTS``.
    """
    return LinePlot(ds, x, y, z, y_err=y_err, x_err=x_err, **plot_opts)


class AutoLinePlot(LinePlot):
    def __init__(self, x, y_z, **lineplot_opts):
        ds = auto_xyz_ds(x, y_z)
        super().__init__(ds, "x", "y", z="z", **lineplot_opts)


def auto_lineplot(x, y_z, **lineplot_opts):
    """Auto version of :func:`~xyzpy.lineplot` that accepts array arguments
    by converting them to a ``Dataset`` first.
    """
    return AutoLinePlot(x, y_z, **lineplot_opts)()


# --------------------------------------------------------------------------- #

_SCATTER_ALT_DEFAULTS = (("legend_handlelength", 0),)


class Scatter(PlotterMatplotlib, AbstractScatter):
    def __init__(self, ds, x, y, z=None, **kwargs):
        # set some scatter specific options
        for k, default in _SCATTER_ALT_DEFAULTS:
            if k not in kwargs:
                kwargs[k] = default
        super().__init__(ds, x, y, z, **kwargs)

    def plot_scatter(self):
        """ """
        self._legend_handles = []
        self._legend_labels = []

        for data in self._gen_xy():
            if "c" in data:
                col = data["c"]
            else:
                col = [next(self._cols)]

            scatter_opts = {
                "c": col,
                "marker": next(self._mrkrs),
                "s": self._marker_size,
                "alpha": self.marker_alpha,
                "label": next(self._zlbls),
                "zorder": next(self._zordrs),
                "rasterized": self.rasterize,
            }

            if "c" in data:
                scatter_opts["cmap"] = self.cmap

            self._legend_handles.append(
                self._axes.scatter(data["x"], data["y"], **scatter_opts)
            )
            self._legend_labels.append(scatter_opts["label"])

    def __call__(self):
        self.prepare_data_single()
        # matplotlib preparation
        self.prepare_plot()
        self.plot_scatter()
        self.plot_legend()
        self.plot_colorbar()
        self.set_panel_label()
        return self.show()


@mpl_multi_plot
@intercept_call_arg
def scatter(ds, x, y, z=None, y_err=None, x_err=None, **plot_opts):
    """From  ``ds`` plot a scatter of ``y`` against ``x``, optionally for
    varying ``z``.

    Parameters
    ----------
    ds : xarray.Dataset
        Dataset to plot from.
    x : str
        Quantity to plot along the x-axis.
    y : str or tuple[str]
        Quantity(s) to plot along the y-axis. If tuple, plot each of the
        variables - instead of ``z``.
    z : str, optional
        Dimension to plot into the page.
    y_err : str, optional
        Variable to plot as y-error.
    x_err : str, optional
        Variable to plot as x-error.
    row : str, optional
        Dimension to vary over as a function of rows.
    col : str, optional
        Dimension to vary over as a function of columns.
    plot_opts
        See ``xyzpy.plot.core.PLOTTER_DEFAULTS``.
    """
    return Scatter(ds, x, y, z, y_err=y_err, x_err=x_err, **plot_opts)


class AutoScatter(Scatter):
    def __init__(self, x, y_z, **scatter_opts):
        ds = auto_xyz_ds(x, y_z)
        super().__init__(ds, "x", "y", z="z", **scatter_opts)


def auto_scatter(x, y_z, **scatter_opts):
    """Auto version of :func:`~xyzpy.scatter` that accepts array arguments
    by converting them to a ``Dataset`` first.
    """
    return AutoScatter(x, y_z, **scatter_opts)()


# --------------------------------------------------------------------------- #

_HISTOGRAM_SPECIFIC_OPTIONS = {
    "stacked": False,
}

_HISTOGRAM_ALT_DEFAULTS = {
    "xtitle": "x",
    "ytitle": "f(x)",
}


class Histogram(PlotterMatplotlib, AbstractHistogram):
    def __init__(self, ds, x, z=None, **kwargs):
        # Set the alternative defaults
        for opt, default in _HISTOGRAM_ALT_DEFAULTS.items():
            if opt not in kwargs:
                kwargs[opt] = default

        # Set histogram specfic options
        for opt, default in _HISTOGRAM_SPECIFIC_OPTIONS.items():
            setattr(self, opt, kwargs.pop(opt, default))

        super().__init__(ds, x, None, z=z, **kwargs)

    def plot_histogram(self):
        from matplotlib.patches import Polygon, Rectangle

        def gen_ind_plots():
            for data in self._gen_xy():
                col = next(self._cols)

                edgecolor = col[:3] + (self.marker_alpha * col[3],)
                facecolor = col[:3] + (self.marker_alpha * col[3] / 4,)
                linewidth = next(self._lws)
                zorder = next(self._zordrs)
                label = next(self._zlbls)

                handle = Rectangle((0, 0), 1, 1, color=facecolor, ec=edgecolor)

                yield (
                    data["x"],
                    edgecolor,
                    facecolor,
                    linewidth,
                    zorder,
                    label,
                    handle,
                )

        xs, ecs, fcs, lws, zds, lbs, hnds = zip(*gen_ind_plots())

        histogram_opts = {
            "label": lbs,
            "bins": self.bins,
            "density": True,
            "histtype": "stepfilled",
            "fill": True,
            "stacked": self.stacked,
            "rasterized": self.rasterize,
        }

        _, _, patches = self._axes.hist(xs, **histogram_opts)

        # Need to set varying colors, linewidths etc seperately
        for patch, ec, fc, lw, zd in zip(patches, ecs, fcs, lws, zds):
            # patch is not iterable if only one set of data created
            if isinstance(patch, Polygon):
                patch = (patch,)

            for sub_patch in patch:
                sub_patch.set_edgecolor(ec)
                sub_patch.set_facecolor(fc)
                sub_patch.set_linewidth(lw)
                sub_patch.set_zorder(zd)

        # store handles for legend
        self._legend_handles, self._legend_labels = hnds, lbs

    def __call__(self):
        # Core preparation
        self.prepare_data_single()
        # matplotlib preparation
        self.prepare_plot()
        self.plot_histogram()
        self.plot_legend()
        self.plot_colorbar()
        self.set_panel_label()
        return self.show()


@mpl_multi_plot
@intercept_call_arg
def histogram(ds, x, z=None, **plot_opts):
    """Dataset histogram.

    Parameters
    ----------
    ds : xarray.Dataset
        The dataset to plot.
    x : str, sequence of str
        The variable(s) to plot the probability density of. If sequence, plot a
        histogram of each instead of using a ``z`` coordinate.
    z : str, optional
        If given, range over this coordinate a plot a histogram for each.
    row : str, optional
        Dimension to vary over as a function of rows.
    col : str, optional
        Dimension to vary over as a function of columns.
    plot_opts
        See ``xyzpy.plot.core.PLOTTER_DEFAULTS``.
    """
    return Histogram(ds, x, z=z, **plot_opts)


class AutoHistogram(Histogram):
    def __init__(self, x, **histogram_opts):
        ds = auto_xyz_ds(x)
        super().__init__(ds, "x", **histogram_opts)


def auto_histogram(x, **histogram_opts):
    """Auto version of :func:`~xyzpy.histogram` that accepts array arguments
    by converting them to a ``Dataset`` first.
    """
    return AutoHistogram(x, **histogram_opts)()


# --------------------------------------------------------------------------- #

_HEATMAP_ALT_DEFAULTS = (
    ("legend", False),
    ("colorbar", True),
    ("colormap", "inferno"),
    ("method", "pcolormesh"),
    ("gridlines", False),
    ("rasterize", True),
)


class HeatMap(PlotterMatplotlib, AbstractHeatMap):
    def __init__(self, ds, x, y, z, **kwargs):
        # set some heatmap specific options
        for k, default in _HEATMAP_ALT_DEFAULTS:
            if k not in kwargs:
                kwargs[k] = default
        super().__init__(ds, x, y, z, **kwargs)

    def plot_heatmap(self):
        """Plot the data as a heatmap."""
        self.calc_color_norm()

        # add extra coords since they *bound* the quads placed -> want ticks
        #     at center of quads
        X = self._heatmap_x
        av_x_bin = np.mean(np.abs(X[:-1] - X[1:]))
        X = np.append(X - av_x_bin / 2, X[-1] + av_x_bin / 2)

        Y = self._heatmap_y
        av_Y_bin = np.mean(np.abs(Y[:-1] - Y[1:]))
        Y = np.append(Y - av_Y_bin / 2, Y[-1] + av_Y_bin / 2)

        self._heatmap = getattr(self._axes, self.method)(
            X,
            Y,
            self._heatmap_var,
            norm=self._color_norm,
            cmap=xyz_colormaps(self.colormap),
            rasterized=self.rasterize,
        )

    def __call__(self):
        # Core preparation
        self.prepare_data_single()
        # matplotlib preparation
        self.prepare_plot()
        self.plot_heatmap()
        self.plot_colorbar()
        self.set_panel_label()
        return self.show()


@mpl_multi_plot
@intercept_call_arg
def heatmap(ds, x, y, z, **kwargs):
    """From  ``ds`` plot variable ``z`` as a function of ``x`` and ``y`` using
    a 2D heatmap.

    Parameters
    ----------
    ds : xarray.Dataset
        Dataset to plot from.
    x : str
        Dimension to plot along the x-axis.
    y : str
        Dimension to plot along the y-axis.
    z : str, optional
        Variable to plot as colormap.
    row : str, optional
        Dimension to vary over as a function of rows.
    col : str, optional
        Dimension to vary over as a function of columns.
    plot_opts
        See ``xyzpy.plot.core.PLOTTER_DEFAULTS``.
    """
    return HeatMap(ds, x, y, z, **kwargs)


class AutoHeatMap(HeatMap):
    def __init__(self, x, **heatmap_opts):
        ds = auto_xyz_ds(x)
        super().__init__(ds, "y", "z", "x", **heatmap_opts)


def auto_heatmap(x, **heatmap_opts):
    """Auto version of :func:`~xyzpy.heatmap` that accepts array arguments
    by converting them to a ``Dataset`` first.
    """
    return AutoHeatMap(x, **heatmap_opts)()


# --------------- Miscellenous matplotlib plotting functions ---------------- #


def setup_fig_ax(
    nrows=1,
    ncols=1,
    facecolor=None,
    rasterize=False,
    rasterize_dpi=300,
    figsize=(5, 5),
    ax=None,
    **kwargs,
):
    import matplotlib.pyplot as plt

    if ax is None:
        fig, ax = plt.subplots(nrows, ncols, figsize=figsize, **kwargs)

        fig.patch.set_alpha(0.0)
    else:
        fig = None

    if not isinstance(ax, np.ndarray):
        # squeezed to single axis
        ax.set_aspect("equal")
        ax.axis("off")

    if facecolor is not None:
        fig.patch.set_facecolor(facecolor)

    if rasterize:
        ax.set_rasterization_zorder(0)
        if fig is not None:
            fig.set_dpi(rasterize_dpi)

    return fig, ax


def show_and_close(fn):
    @functools.wraps(fn)
    def wrapped(*args, show_and_close=True, **kwargs):
        import warnings

        import matplotlib.pyplot as plt

        # remove annoying regular warning about all-nan slices
        with warnings.catch_warnings():
            warnings.filterwarnings(
                action="ignore",
                message="All-NaN slice",
            )
            fig, ax = fn(*args, **kwargs)

        if fig is not None:
            if show_and_close:
                plt.show()
                plt.close(fig)

        return fig, ax

    return wrapped


def choose_squarest_grid(x):
    p = x**0.5
    if p.is_integer():
        m = n = int(p)
    else:
        m = int(round(p))
        p = int(p)
        n = p if m * p >= x else p + 1
    return m, n


def _compute_hue(z):
    # the constant rotation is to have blue and orange as + and -
    # negation puts +i as green -i as pink
    return ((-np.angle(z) + 9 * np.pi / 8) / (2 * np.pi)) % 1


def to_colors(
    zs,
    magscale="linear",
    max_mag=None,
    alpha_map=True,
    alpha_pow=1 / 2,
):
    """Convert complex array to RGBA colors.

    Parameters
    ----------
    zs : array_like
        Complex array to convert to colors.
    magscale : {"linear", "log", float}, optional
        Scale to apply to the magnitude before mapping to saturation.
        If "linear", no scaling is applied. If "log", logarithmic scaling
        is applied. If float, the magnitude is raised to this power.
        Default is "linear".
    max_mag : float, optional
        Maximum magnitude to map to full saturation. If None, the maximum
        magnitude in `zs` is used.
    alpha_map : bool, optional
        Whether to append an alpha channel based on magnitude. Default is True.
    alpha_pow : float, optional
        Power to raise the magnitude to when computing the alpha channel.
        Default is 1/2.

    Returns
    -------
    colors : ndarray
        Array of shape (..., 3) or (..., 4) of RGB(A) colors.
    mapped_mag : ndarray
        The mapped magnitudes used for saturation and alpha.
    max_mag : float
        The maximum magnitude used for saturation mapping.
    """
    import matplotlib as mpl

    arraymag = np.abs(zs)

    if magscale == "linear":
        mapped_mag = arraymag
    elif magscale == "log":
        # XXX: need some kind of cutoff?
        raise NotImplementedError("log scale not implemented.")
    else:
        # more robust way to 'flatten' the perceived magnitude
        mapped_mag = arraymag**magscale

    if max_mag is None:
        max_mag = np.max(mapped_mag)

    hue = _compute_hue(zs)
    sat = mapped_mag / max_mag
    val = np.tile(1.0, hue.shape)
    zs = mpl.colors.hsv_to_rgb(np.stack((hue, sat, val), axis=-1))

    if alpha_map:
        # append alpha channel
        zalpha = (mapped_mag / max_mag) ** alpha_pow
        zs = np.concatenate([zs, np.expand_dims(zalpha, -1)], axis=-1)

    return zs, mapped_mag, max_mag


def add_visualize_legend(
    ax,
    complexobj,
    max_mag,
    legend_inside=False,
    auto_pad=0.03,
    legend_loc="auto",
    legend_size=0.15,
    legend_bounds=None,
    legend_resolution=3,
):
    import matplotlib as mpl

    # choose where to put the legend
    if legend_bounds is None:
        if legend_loc in ("auto", "bottom right"):
            if not legend_inside:
                # move compass and legends beyond the plot rectangle, which
                # will be filled when there are only 2 plot dimensions
                legend_loc = (1 - auto_pad, 0.0 - legend_size + auto_pad)
            else:
                # occupy space within the rectangle
                legend_loc = (1.0 - legend_size, 0.0)

        elif legend_loc == "top right":
            if not legend_inside:
                legend_loc = (1 - auto_pad, 1 - legend_size - auto_pad)
            else:
                legend_loc = (1.0 - legend_size, 1.0 - legend_size)

        legend_bounds = (*legend_loc, legend_size, legend_size)

    lax = ax.inset_axes(legend_bounds)
    lax.axis("off")
    lax.set_aspect("equal")

    num = legend_resolution * 2 + 1
    if complexobj:
        re = np.linspace(-1, 1, num=num).reshape(1, -1)
        im = 1j * np.linspace(1, -1, num=num).reshape(-1, 1)
        z = re + im
    else:
        re = np.linspace(-1, 1, num=num)
        # repeat a few times to make the bar thick enough
        z = np.tile(re, (max(1, legend_resolution // 3), 1))

    zmag = np.abs(z)
    mask = zmag > 1.0
    z[mask] = zmag[mask] = 0.0

    # compute the color for each point
    hue = _compute_hue(z)
    sat = zmag
    val = np.tile(1.0, hue.shape)

    # convert to rgb
    bars = mpl.colors.hsv_to_rgb(np.stack((hue, sat, val), axis=-1))
    # add alpha channel
    bars = np.concatenate([bars, np.tile(1.0, hue.shape + (1,))], -1)
    # add make area outside disk transparent
    bars[..., 3][mask] = 0.0

    # plot the actual colorbar legend
    lax.imshow(bars)

    # add axis orientation labels
    lax.text(
        1.02,
        0.5,
        "$+1$",
        ha="left",
        va="center",
        transform=lax.transAxes,
        color=(0.5, 0.5, 0.5),
        size=6,
    )
    lax.text(
        -0.02,
        0.5,
        "$-1$",
        ha="right",
        va="center",
        transform=lax.transAxes,
        color=(0.5, 0.5, 0.5),
        size=6,
    )
    if complexobj:
        lax.text(
            0.5,
            -0.02,
            "$-j$",
            ha="center",
            va="top",
            transform=lax.transAxes,
            color=(0.5, 0.5, 0.5),
            size=6,
        )
        lax.text(
            0.5,
            1.02,
            "$+j$",
            ha="center",
            va="bottom",
            transform=lax.transAxes,
            color=(0.5, 0.5, 0.5),
            size=6,
        )

    # show the overall scale
    if complexobj:
        overall_scale_opts = {"x": 0.85, "y": 0.15, "ha": "left"}
    else:
        overall_scale_opts = {"x": 0.5, "y": -0.15, "ha": "center"}

    lax.text(
        s=f"$\\times {float(max_mag):.3}$",
        va="top",
        **overall_scale_opts,
        transform=lax.transAxes,
        color=(0.5, 0.5, 0.5),
        size=6,
    )


def make_ax_square_after_plotted(ax):
    xmin, xmax = ax.get_xlim()
    xrange = abs(xmax - xmin)
    ymin, ymax = ax.get_ylim()
    yrange = abs(ymax - ymin)

    # pad either x or y to make square
    if xrange > yrange:
        ypad = (xrange - yrange) / 2
        if ymin > ymax:
            # fipped y axis
            ax.set_ylim(ymin + ypad, ymax - ypad)
        else:
            ax.set_ylim(ymin - ypad, ymax + ypad)

    elif yrange > xrange:
        xpad = (yrange - xrange) / 2
        if xmin > xmax:
            # flipped x axis
            ax.set_xlim(xmin + xpad, xmax - xpad)
        else:
            ax.set_xlim(xmin - xpad, xmax + xpad)


def handle_sequence_of_arrays(f):
    """Simple wrapper to handle sequence of arrays as input to e.g.
    ``visualize_tensor``.
    """

    @functools.wraps(f)
    def wrapped(array, *args, show_and_close=True, **kwargs):
        import matplotlib.pyplot as plt

        if isinstance(array, (tuple, list)) and all(
            hasattr(x, "shape") for x in array
        ):
            # assume sequence of tensors to plot
            figsize = kwargs.get("figsize", (5, 5))
            rasterize_dpi = kwargs.get("rasterize_dpi", 300)
            nplot = len(array)
            fig, axs = plt.subplots(
                1,
                nplot,
                figsize=figsize,
                squeeze=False,
                sharey=True,
            )
            fig.set_dpi(rasterize_dpi)
            fig.patch.set_alpha(0.0)

            # plot all tensors with same magnitude scale
            kwargs.setdefault("max_mag", max(np.max(np.abs(x)) for x in array))

            # only show legend on last plot
            legend = kwargs.pop("legend", True)

            for i in range(nplot):
                f(
                    array[i],
                    *args,
                    ax=axs[0, i],
                    # only show legend on last plot
                    legend=(legend and i == nplot - 1),
                    show_and_close=False,
                    **kwargs,
                )
                make_ax_square_after_plotted(axs[0, i])

            if show_and_close:
                plt.show()
                plt.close(fig)
            return fig, axs

        else:
            # treat as single tensor
            return f(array, *args, show_and_close=show_and_close, **kwargs)

    return wrapped


@handle_sequence_of_arrays
@show_and_close
def visualize_matrix(
    array,
    max_mag=None,
    magscale="linear",
    alpha_map=True,
    alpha_pow=1 / 2,
    legend=True,
    legend_loc="auto",
    legend_size=0.15,
    legend_bounds=None,
    legend_resolution=3,
    facecolor=None,
    rasterize=4096,
    rasterize_dpi=300,
    figsize=(5, 5),
    ax=None,
):
    """Visualize ``array`` as a 2D colormapped image.

    Parameters
    ----------
    array : array_like or Sequence[array_like]
        A 2D (or 1D) array or sequence of arrays to visualize.
    max_mag : float, optional
        The maximum magnitude to use for the color mapping. If not provided,
        the maximum magnitude in the array will be used.
    magscale : "linear" or float, optional
        How to scale the magnitude of the array values. If "linear", then
        the magnitude is used directly. If a float, then the magnitude is
        raised to this power before being used, which can help to show
        variation among small values.
    alpha_map : bool, optional
        Whether to map the tensor value magnitudes to pixel alpha.
    alpha_pow : float, optional
        The power to raise the magnitude to when mapping to alpha.
    legend : bool, optional
        Whether to show a legend (colorbar). If the array has complex dtype
        then the legend will be a colorwheel.
    legend_loc : str or tuple[float], optional
        Where to place the legend. If "auto", then the legend will be placed
        outside the plot rectangle, otherwise it should be a tuple of
        ``(x, y)`` coordinates in axes space.
    legend_size : float, optional
        The size of the legend, in relation to the size of the plot axes.
    legend_bounds : tuple[float], optional
        The bounds of the legend, as ``(x, y, width, height)`` in axes space.
        If not provided, the bounds will be computed from ``legend_loc`` and
        ``legend_size``.
    legend_resolution : int, optional
        The number of different colors to show in the legend.
    facecolor : str, optional
        The background color of the plot, by default transparent.
    rasterize : int or float, optional
        Whether to rasterize the plot. If a number, then rasterize if the
        number of pixels in the plot is greater than this value.
    rasterize_dpi : float, optional
        The dpi to use when rasterizing.
    figsize : tuple[float], optional
        The size of the figure to create, if ``ax`` is not provided.
    ax : matplotlib.Axis, optional
        The axis to draw to. If not provided, a new figure will be created.
    show_and_close : bool, optional
        If ``True`` (the default) then show and close the figure, otherwise
        return the figure and axis.

    Returns
    -------
    fig : matplotlib.Figure
        The figure containing the plot, or ``None`` if ``ax`` was provided.
    ax : matplotlib.Axis
        The axis or axes containing the plot(s).
    """
    # can only plot numpy
    array = np.asarray(array)
    if array.ndim == 1:
        # draw vectors as diagonals
        array = np.diag(array)

    if isinstance(rasterize, (float, int)):
        # only turn on above a certain size
        rasterize = array.size > rasterize

    fig, ax = setup_fig_ax(
        facecolor=facecolor,
        rasterize=rasterize,
        rasterize_dpi=rasterize_dpi,
        figsize=figsize,
        ax=ax,
    )

    zs, _, max_mag = to_colors(
        array,
        magscale=magscale,
        max_mag=max_mag,
        alpha_map=alpha_map,
        alpha_pow=alpha_pow,
    )
    ax.imshow(zs, interpolation="nearest", zorder=-1)

    if legend:
        add_visualize_legend(
            ax=ax,
            complexobj=np.iscomplexobj(array),
            max_mag=max_mag,
            legend_inside=False,
            legend_loc=legend_loc,
            legend_size=legend_size,
            legend_bounds=legend_bounds,
            legend_resolution=legend_resolution,
        )

    return fig, ax


@handle_sequence_of_arrays
@show_and_close
@autocorrect_kwargs
def visualize_tensor(
    array,
    spacing_factor=1.0,
    max_projections=None,
    projection_overlap_spacing=1.05,
    angles=None,
    scales=None,
    skew_angle_factor="auto",
    skew_scale_factor=0.05,
    max_mag=None,
    magscale="linear",
    size_map=True,
    size_pow=1 / 2,
    size_scale=1.0,
    alpha_map=True,
    alpha_pow=1 / 2,
    alpha=0.8,
    marker="o",
    linewidths=0,
    show_lattice=True,
    lattice_opts=None,
    compass=False,
    compass_loc="auto",
    compass_size=0.1,
    compass_bounds=None,
    compass_labels=None,
    compass_opts=None,
    legend=True,
    legend_loc="auto",
    legend_size=0.15,
    legend_bounds=None,
    legend_resolution=3,
    interleave_projections=False,
    reverse_projections=False,
    facecolor=None,
    rasterize=4096,
    rasterize_dpi=300,
    figsize=(5, 5),
    ax=None,
):
    """Visualize all entries of a tensor, with indices mapped into the plane
    and values mapped into a color wheel.

    Parameters
    ----------
    array : numpy.ndarray
        The tensor to visualize.
    spacing_factor : float, optional
        How to scale the dimensions relative to each other. If 1.0, then each
        dimension will have the same extent, and smaller dimensions will be
        sparser. If 0.0, the each dimension will have an extent propoertional
        to its size, with matching density.
    max_projections : int, optional
        The maximum number of different projection directions / angles to use.
        If specified and less than the number of dimensions, then multiple
        dimensions will share the same angle but with different scales.
    projection_overlap_spacing : float, optional
        When grouping multiple dimensions to the same angle, how much to
        increase the spacing at each scale so as to emphasize each.
    angles : sequence[float], optional
        An explicit list of angles to use for each direction, in radians, with
        zero pointing straight down. If not provided, then the angles will be
        calculated automatically.
    scales : sequence[float], optional
        An explicit list of scales to use for each direction. If not provided,
        then the scales will be calculated automatically.
    skew_angle_factor : float, optional
        When there are more than two dimensions, a factor to scale the
        rotations by to avoid overlapping data points. If 0.0 then the angles
        will be evenly spaced.
    skew_scale_factor : float, optional
        When there are more than two dimensions, a factor to scale the
        scales by to avoid overlapping data points, that shortens
        non-perpendicular directions.
    max_mag : float, optional
        The maximum magnitude to use for the color mapping. If not provided,
        the maximum magnitude in the array will be used.
    magscale : "linear" or float, optional
        How to scale the magnitude of the array values. If "linear", then
        the magnitude is used directly. If a float, then the magnitude is
        raised to this power before being used, which can help to show
        variation among small values.
    size_map : bool, optional
        Whether to map the tensor value magnitudes to marker size.
    size_scale : float, optional
        An overall factor to scale the marker size by.
    alpha_map : bool, optional
        Whether to map the tensor value magnitudes to marker alpha.
    alpha_pow : float, optional
        The power to raise the magnitude to when mapping to alpha.
    alpha : float, optional
        The overall alpha to use for all markers if ``not alpha_map``.
    marker : str, optional
        The marker to use for the markers.
    linewidths : float, optional
        The linewidth to use for the markers.
    show_lattice : bool, optional
        Show a thin grey line connecting adjacent array coordinate points.
    lattice_opts : dict, optional
        Options to pass to ``maplotlib.Axis.scatter`` for the lattice grid.
    compass : bool, optional
        Whether to show a compass indicating the orientation of each dimension.
    compass_loc : (float, float), optional
        Where to place the compass.
    compass_size : float, optional
        The size of the compass.
    compass_bounds : tuple[float], optional
        Explicit bounds of the compass, as ``(x, y, width, height)``.
    compass_labels : sequence[str], optional
        Explicit labels for the compass, in order of the dimensions.
    compass_opts : dict, optional
        Extra options for the compass arrows.
    legend : bool, optional
        Whether to show a legend (colorbar). If the array has complex dtype
        then the legend will be a colorwheel.
    legend_loc : str or tuple[float], optional
        Where to place the legend. If "auto", then the legend will be placed
        outside the plot rectangle, otherwise it should be a tuple of
        ``(x, y)`` coordinates in axes space.
    legend_size : float, optional
        The size of the legend, in relation to the size of the plot axes.
    legend_bounds : tuple[float], optional
        Explicit bounds of the legend, as ``(x, y, width, height)`` in axes
        space.
    legend_resolution : int, optional
        The number of different colors to show in the legend.
    interleave_projections : bool, optional
        If ``True`` and grouping dimensions, then they are assigned round robin
        fashion rather than blocks. ``False`` matches the behavior of fusing.
    reverse_projections : bool, optional
        Whether to reverse the order of the projections.
    facecolor : str, optional
        The background color of the plot, by default transparent.
    rasterize : int or float, optional
        Whether to rasterize the plot. If a number, then rasterize if the
        size of the array is greater than this value.
    rasterize_dpi : float, optional
        The dpi to use when rasterizing.
    figsize : tuple, optional
        The size of the figure to create, if ``ax`` is not provided.
    ax : matplotlib.Axis, optional
        The axis to draw to. If not provided, a new figure will be created.

    Returns
    -------
    fig : matplotlib.Figure
        The figure containing the plot, or ``None`` if ``ax`` was provided.
    ax : matplotlib.Axis
        The axis containing the plot.
    """
    import matplotlib as mpl

    # can only plot numpy
    array = np.asarray(array)

    if isinstance(rasterize, (float, int)):
        # only turn on above a certain size
        rasterize = array.size > rasterize

    fig, ax = setup_fig_ax(
        facecolor=facecolor,
        rasterize=rasterize,
        rasterize_dpi=rasterize_dpi,
        figsize=figsize,
        ax=ax,
    )

    if max_projections is None:
        max_projections = array.ndim

    auto_skew = skew_angle_factor == "auto"
    if auto_skew:
        if max_projections == 3:
            skew_angle_factor = 0.3
        else:
            skew_angle_factor = 0.05

    auto_angles = angles is None
    if scales == "equal":
        scales = [1] * array.ndim

    auto_scales = scales is None

    # map each dimension to an angle
    if not auto_angles:
        angles = np.array(angles)
    else:
        # if max_projections == array.ndim, then each dimension has its own
        # angle, if max_projections < array.dim, then we will
        # reuse the same angles, initially round robin distributed
        angles = np.tile(
            np.linspace(0.0, np.pi, max_projections, endpoint=False),
            array.ndim // max_projections + 1,
        )[: array.ndim]

        if not interleave_projections:
            # 'fill up' one angle before moving on, rather than round-robin,
            # doing this matches the behavior of fusing adjacent dimensions
            angles = np.sort(angles)

        def angle_modulate(x):
            return x * (x - np.pi / 2) * (x - x[-1])

        # modulate the angles slightly to avoid overlapping data points
        angles += angle_modulate(angles) * skew_angle_factor

    if auto_scales:
        scales = np.empty(angles.shape)
    else:
        scales = np.array(scales)

    # the logic here is, when grouping dimensions into the same angles we
    # need to offset each overlapping dimension by increasing amount
    first_size = {}
    grouped_size = {}
    group_counter = {}
    group_rank = {}
    fastest_varying = []

    iphis = list(enumerate(angles))
    if not reverse_projections:
        iphis.reverse()

    for i, phi in iphis:
        if phi not in first_size:
            # first time we have encountered an axis at this angle
            fastest_varying.append((i, array.shape[i]))
            first_size[phi] = array.shape[i]
            grouped_size[phi] = array.shape[i]
            group_counter[phi] = 1
        else:
            # already an axis at this angle, space this one larger
            grouped_size[phi] *= array.shape[i]
            group_counter[phi] += 1

        # what rank among axes at this angle is i?
        group_rank[i] = group_counter[phi]

        if auto_scales:
            scales[i] = (
                # base size
                (grouped_size[phi] // array.shape[i])
                # shrink non-perpendicular dimensions
                * (1 + skew_scale_factor * np.cos(4 * angles[i]))
                # put extra space between distinct dimensions
                * projection_overlap_spacing ** group_counter[phi]
                # how much to match the first size by
                * max(1, (first_size[phi] - 1)) ** -spacing_factor
            )

    eff_width = max(grouped_size.values())
    eff_ndim = max_projections

    # define the core mappings of coordinate to 2D plane

    def xcomponent(i, coo):
        return scales[i] * np.sin(angles[i]) * coo

    def ycomponent(i, coo):
        return scales[i] * -np.cos(angles[i]) * coo

    # compute projection into 2D coordinates for every index location
    coos = np.indices(array.shape)
    x = np.zeros(array.shape)
    y = np.zeros(array.shape)
    for i, coo in enumerate(coos):
        x += xcomponent(i, coo)
        y += ycomponent(i, coo)

    # compute colors
    zs, mapped_mag, max_mag = to_colors(
        array.flat,
        magscale=magscale,
        max_mag=max_mag,
        alpha_map=alpha_map,
        alpha_pow=alpha_pow,
    )

    # compute a sensible base size based on expected density of points
    base_size = size_scale * 3000 / (eff_width * eff_ndim**1.2)
    if size_map:
        s = base_size * (mapped_mag / max_mag) ** size_pow
    else:
        s = base_size

    if show_lattice:
        # put a small grey line on every edge
        ls = []
        for i, isize in fastest_varying:
            other_shape = array.shape[:i] + array.shape[i + 1 :]
            other_coos = np.indices(other_shape)
            coo_start = np.insert(other_coos, i, 0, 0)
            coo_stop = np.insert(other_coos, i, isize - 1, 0)

            xi = np.zeros(coo_start.shape)
            yi = np.zeros(coo_start.shape)

            for i, coo in enumerate(coo_start):
                xi += xcomponent(i, coo)
                yi += ycomponent(i, coo)

            xf = np.zeros(coo_start.shape)
            yf = np.zeros(coo_start.shape)

            for i, coo in enumerate(coo_stop):
                xf += xcomponent(i, coo)
                yf += ycomponent(i, coo)

            li = np.stack((xi.flat, yi.flat), 1)
            lf = np.stack((xf.flat, yf.flat), 1)

            ls.append(np.stack((li, lf), 1))

        segments = np.concatenate(ls)

        lattice_opts = {} if lattice_opts is None else dict(lattice_opts)
        lattice_opts.setdefault("color", (0.6, 0.6, 0.6))
        lattice_opts.setdefault(
            "alpha", 0.01 + 2 ** (-(array.size**0.2 + eff_ndim**0.8))
        )
        lattice_opts.setdefault("linewidth", 1)
        lattice_opts.setdefault("zorder", -2)
        lines = mpl.collections.LineCollection(segments, **lattice_opts)
        ax.add_collection(lines)

    # plot the actual points
    ax.scatter(
        # mapped variables
        # (reverse the data so that the correct points are shown on top)
        x=x.flat[::-1],
        y=y.flat[::-1],
        c=zs[::-1],
        s=s[::-1] if size_map else s,
        # constants
        alpha=None if alpha_map else alpha,  # folded into color if alpha_map
        linewidths=linewidths,
        marker=marker,
        zorder=-1,
        clip_on=False,
    )

    if compass:
        # choose where to put the compass
        if compass_bounds is None:
            if compass_loc == "auto":
                if max_projections <= 2:
                    # move compass and legends beyond the plot rectangle, which
                    # will be filled when there are only 2 plot dimensions
                    compass_loc = (-0.05, 1.0 - compass_size + 0.05)
                else:
                    # occupy space within the rectangle
                    compass_loc = (0.0, 1 - compass_size)
            compass_bounds = (*compass_loc, compass_size, compass_size)

        cax = ax.inset_axes(compass_bounds)
        cax.axis("off")
        cax.set_aspect("equal")

        compass_opts = {} if compass_opts is None else dict(compass_opts)
        compass_opts.setdefault("color", (0.5, 0.5, 0.5))
        compass_opts.setdefault("width", 0.002)
        compass_opts.setdefault("length_includes_head", True)

        if compass_labels is None:
            compass_labels = range(len(angles))
        elif compass_labels is False:
            compass_labels = [""] * len(angles)

        for i, phi in enumerate(angles):
            modifier = 1 + skew_scale_factor * np.cos(4 * angles[i])
            dx = np.sin(phi) * group_rank[i] * modifier
            dy = -np.cos(phi) * group_rank[i] * modifier
            cax.arrow(0, 0, dx, dy, **compass_opts)
            cax.text(
                dx,
                dy,
                f" {compass_labels[i]}",
                ha="left",
                va="top",
                color=compass_opts["color"],
                size=6,
                rotation=180 * phi / np.pi - 90,
                rotation_mode="anchor",
            )

    if legend:
        add_visualize_legend(
            ax=ax,
            complexobj=np.iscomplexobj(array),
            max_mag=max_mag,
            legend_inside=max_projections > 2,
            legend_loc=legend_loc,
            legend_size=legend_size,
            legend_bounds=legend_bounds,
            legend_resolution=legend_resolution,
        )

    return fig, ax


def _draw_tensor_tree_branch(
    lines,
    widths,
    colors,
    array,
    x=0.0,
    y=0.0,
    phi_center=np.pi / 2,
    phi_range=np.pi,
    level=1,
    z=None,
    mag=None,
    mag_weight=None,
    linewidth=10,
):
    if not array.shape:
        # stop drawing
        return

    d, *subshape = array.shape
    phis = np.linspace(-0.5, 0.5, 2 * d + 1)[1:-1:2]

    for i in range(d):
        subarray = array[i]
        subphi = phi_center + phi_range * phis[i]

        subx = -level * np.cos(subphi)
        suby = level * np.sin(subphi)

        subz = z[i]
        submag = mag[i]

        _draw_tensor_tree_branch(
            lines,
            widths,
            colors,
            subarray,
            x=subx,
            y=suby,
            phi_center=subphi,
            phi_range=(phi_range / d),
            level=level + 1,
            z=subz,
            mag=submag,
            mag_weight=mag_weight,
            linewidth=linewidth,
        )

        width = linewidth * (np.sum(submag) / mag_weight) ** 0.5

        if not subshape:
            color = subz
        else:
            color = np.mean(subz, axis=tuple(range(subz.ndim - 1)))

        lines.append([(x, y), (subx, suby)])
        widths.append(width)
        colors.append(color)


def visualize_tensor_tree(
    array,
    phi_center=np.pi / 2,
    phi_range=np.pi,
    linewidth=10,
    legend=True,
    legend_loc="top right",
    legend_size=0.15,
    legend_bounds=None,
    legend_resolution=3,
    facecolor=None,
    rasterize=4096,
    rasterize_dpi=300,
    figsize=(5, 5),
    ax=None,
):
    from matplotlib.collections import LineCollection

    z, mag, max_mag = to_colors(
        array,
        magscale=0.5,
        alpha_map=False,
    )
    mag_weight = np.sum(mag)

    lines = []
    widths = []
    colors = []

    _draw_tensor_tree_branch(
        lines=lines,
        widths=widths,
        colors=colors,
        array=array,
        z=z,
        mag=mag,
        mag_weight=mag_weight,
        phi_center=phi_center,
        phi_range=phi_range,
        linewidth=linewidth,
    )

    fig, ax = setup_fig_ax(
        facecolor=facecolor,
        rasterize=rasterize,
        rasterize_dpi=rasterize_dpi,
        figsize=figsize,
        ax=ax,
    )

    lc = LineCollection(
        lines,
        linewidths=widths,
        colors=colors,
        zorder=-1,
        capstyle="round",
    )

    ax.add_collection(lc)
    ax.autoscale()

    if legend:
        add_visualize_legend(
            ax=ax,
            complexobj=np.iscomplexobj(array),
            max_mag=max_mag,
            legend_inside=True,
            legend_loc=legend_loc,
            legend_size=legend_size,
            legend_bounds=legend_bounds,
            legend_resolution=legend_resolution,
        )

    return fig, ax



"""
Helper functions for preparing data to be plotted.
"""
# TODO: x-err, upper and lower errors *************************************** #
# TODO: check shape and hint at which dimensions need to be reduced ********* #

import functools
import itertools

import numpy as np
import numpy.ma as ma
import xarray as xr

from .color import convert_colors, get_default_sequential_cm, xyz_colormaps
from .marker import _MARKERS, _SINGLE_LINE_MARKER

PLOTTER_DEFAULTS = {
    # Figure options
    "backend": "MATPLOTLIB",
    "figsize": (7, 6),
    "axes_loc": None,
    "axes_rloc": None,
    "add_to_axes": None,
    "add_to_xaxes": None,
    "add_to_fig": None,
    "subplot": None,
    "fignum": 1,
    "return_fig": True,
    # Coloring options
    "colors": None,
    "colormap": None,
    "colormap_log": False,
    "colormap_reverse": False,
    # Colorbar options
    "colorbar": None,
    "ctitle": None,
    "vmin": None,
    "vmax": None,
    "colorbar_relative_position": None,
    "colorbar_where": None,
    "colorbar_opts": dict(),
    "method": None,
    "colorbar_color": "black",
    # 'z-axis' options: for legend or colorbar
    "ztitle": None,
    "zlabels": None,
    "zlims": (None, None),
    "zticks": None,
    "legend": None,
    "legend_loc": "best",
    "legend_where": None,
    "legend_ncol": 1,
    "legend_bbox": None,
    "legend_marker_scale": None,
    "legend_label_spacing": None,
    "legend_column_spacing": 1,
    "legend_frame": False,
    "legend_frame_alpha": 0.5,
    "legend_handlelength": None,
    "legend_reverse": False,
    "legend_marker_alpha": None,
    # x-axis options
    "xtitle": None,
    "xtitle_pad": 5,
    "xlims": None,
    "xticks": None,
    "xtick_labels": None,
    "xticklabels_hide": False,
    "xlog": False,
    "xjitter": None,
    "bins": 30,
    # y-axis options
    "ytitle": None,
    "ytitle_pad": 5,
    "ytitle_right": False,
    "ylims": None,
    "yticks": None,
    "yticklabels_hide": False,
    "yticklabels_right": None,
    "ylog": False,
    "yjitter": None,
    # Titles and text
    "title": None,
    "title_pad": None,
    "panel_label": None,
    "panel_label_loc": (0.05, 0.93),
    "panel_label_color": "black",
    # Styling options
    "markers": None,
    "marker_size": None,
    "marker_alpha": 1.0,
    "lines": True,
    "line_styles": None,
    "line_widths": None,
    "errorbar_capthick": 1,
    "errorbar_capsize": 0,
    "errorbar_linewidth": 0.7,
    "zorders": None,
    "padding": None,
    "vlines": None,
    "hlines": None,
    "span_style": "--",
    "span_width": 1,
    "span_color": "0.5",
    "gridlines": True,
    "gridline_style": (1, 2),
    "ticks_where": ("bottom", "left", "top", "right"),
    # Font options
    "font": ("CMU Serif", "PT Serif", "Liberation Serif", "DejaVu Serif"),
    "font_weight": "roman",
    "fontsize_title": 18,
    "fontsize_ticks": 13,
    "fontsize_xtitle": 18,
    "fontsize_ytitle": 18,
    "fontsize_ztitle": 18,
    "fontsize_zlabels": 13,
    "fontsize_panel_label": 15,
    "math_serif": True,
    # backend options
    "rasterize": False,
    "webgl": False,
}

_PLOTTER_OPTS = list(PLOTTER_DEFAULTS.keys())


def check_excess_dims(ds, var, valid_dims, mode="lineplot"):

    if mode == "lineplot":
        expanded_valid_dims = list(
            itertools.chain.from_iterable(ds[d].dims for d in valid_dims)
        )

        excess_dims = {
            d
            for d, sz in ds[var].sizes.items()
            if (sz > 1) and (d not in expanded_valid_dims)
        }
        if excess_dims:
            raise ValueError(
                "Dataset has too many non-singlet dimensions "
                "- try selection values for the following: "
                "{}.".format(excess_dims)
            )

    else:
        # TODO: for now, no checking if not lineplot
        pass


class Plotter:
    """ """

    def __init__(
        self, ds, x, y, z=None, c=None, y_err=None, x_err=None, **kwargs
    ):
        """ """
        if isinstance(ds, xr.DataArray):
            self._ds = ds.to_dataset()
        else:
            self._ds = ds

        self.x_coo = x
        self.y_coo = y
        self.z_coo = z
        self.c_coo = c
        self.y_err = y_err
        self.x_err = x_err

        # Figure options
        settings = {**PLOTTER_DEFAULTS, **kwargs}
        for opt in _PLOTTER_OPTS:
            setattr(self, opt, settings.pop(opt))

        # Parse unmatched keywords and error but suggest correct versions
        if len(settings) > 0:
            import difflib

            wrong_opts = list(settings.keys())
            right_opts = [
                difflib.get_close_matches(opt, _PLOTTER_OPTS, n=3)
                for opt in wrong_opts
            ]

            msg = "Option(s) {} not valid.\n Did you mean: {}?".format(
                wrong_opts, right_opts
            )
            print(msg)
            raise ValueError(msg)

        if self.colors and self.c_coo:
            raise ValueError("Cannot specify explicit colors if ``c`` used.")

        # Internal state
        self._data_range_calculated = False
        self._c_cols = []

    # ----------------------------- properties ------------------------------ #

    def _get_ds(self):
        return self._ds

    def _set_ds(self, ds):
        self._ds = ds
        self.prepare_z_vals()
        self.prepare_z_labels()
        self.prepare_xy_vals_lineplot()
        self.calc_plot_range()
        self.update()

    def _del_ds(self):
        self._ds = None

    ds = property(_get_ds, _set_ds, _del_ds, "The dataset used to plot graph.")

    # ------------------------------- methods ------------------------------- #

    def prepare_axes_labels(self):
        """Work out what the axes titles should be."""
        self._xtitle = self.x_coo if self.xtitle is None else self.xtitle
        if self.ytitle is None:
            if isinstance(self.y_coo, str):
                self._ytitle = self.y_coo
            else:
                self._ytitle = None
        else:
            self._ytitle = self.ytitle

        if self.c_coo is None:
            self._ctitle = self.z_coo if self.ztitle is None else self.ztitle
        else:
            self._ctitle = self.c_coo if self.ctitle is None else self.ctitle

    def prepare_z_vals(self, mode="lineplot", grid=False):
        """Work out what the 'z-coordinate', if any, should be.

        Parameters
        ----------
        mode : {'lineplot', 'scatter', 'histogram'}
        """
        # TODO: process multi-var errors at same time
        self._multi_var = False

        # Multiple sets of data parametized by z_coo
        if self.z_coo is not None:
            if not grid:
                check_excess_dims(
                    self._ds, self.y_coo, (self.x_coo, self.z_coo), mode=mode
                )
            self._z_vals = self._ds[self.z_coo].values

        # Multiple data variables to plot -- just named in list
        elif isinstance(self.y_coo, (tuple, list)):
            for var in self.y_coo:
                if grid:
                    check_excess_dims(self._ds, var, (self.x_coo,), mode=mode)
            self._multi_var = True
            self._z_vals = self.y_coo

        # Multiple data variables to plot -- but for histogram
        elif isinstance(self.x_coo, (tuple, list)):
            self._multi_var = True
            self._z_vals = self.x_coo

        # Single data variable to plot
        else:
            if not grid:
                check_excess_dims(
                    self._ds, self.y_coo, (self.x_coo,), mode=mode
                )
            self._z_vals = (None,)

    def prepare_z_labels(self):
        """Work out what the labels for the z-coordinate should be."""
        # Manually specified z-labels
        if self.zlabels is not None:
            self._zlbls = iter(self.zlabels)

        # Use z-data (for lineplot only) or multiple y names
        elif (self.z_coo is not None) or self._multi_var:
            self._zlbls = iter(str(z) for z in self._z_vals)

        # No z-labels
        else:
            self._zlbls = itertools.repeat(None)

    def calc_use_legend_or_colorbar(self):
        """Work out whether to use a legend."""

        def auto_legend():
            return 1 < len(self._z_vals) <= 10

        # cspecified, lspecified, many points, colors, colors, c_coo
        if self.colorbar and (self.legend is None):
            self.legend = False if (self.c_coo is None) else auto_legend()
        if self.legend and (self.colorbar is None):
            self.colorbar = False if (self.c_coo is None) else True

        if self.legend is None and self.colorbar is None:
            self._use_legend = auto_legend()
            self._use_colorbar = (
                (not self._use_legend) and self.colors is True
            ) or self.c_coo is not None
        else:
            self._use_legend = self.legend
            self._use_colorbar = self.colorbar

    def prepare_xy_vals_lineplot(self, mode="lineplot"):
        """Select and flatten the data appropriately to iterate over."""

        def gen_xy():
            for i, z in enumerate(self._z_vals):
                das = {}
                data = {}

                # multiple data variables rather than z coordinate
                if self._multi_var:
                    das["x"] = self._ds[self.x_coo]
                    das["y"] = self._ds[z]

                    if (
                        (self.y_err is not None)
                        or (self.x_err is not None)
                        or (self.c_coo is not None)
                    ):
                        raise ValueError("Multi-var errors/c not implemented.")

                # z-coordinate to iterate over
                elif z is not None:
                    try:
                        # try positional indexing first, as much faster
                        sub_ds = self._ds[{self.z_coo: i}]
                    except ValueError:
                        # but won't work e.g. on non-dimensions
                        sub_ds = self._ds.loc[{self.z_coo: z}]

                    das["x"] = sub_ds[self.x_coo]
                    das["y"] = sub_ds[self.y_coo]

                    if self.c_coo is not None:
                        if mode == "lineplot":
                            self._c_cols.append(
                                np.asscalar(
                                    sub_ds[self.c_coo].values.flatten()
                                )
                            )
                        elif mode == "scatter":
                            das["c"] = sub_ds[self.c_coo]

                    if self.y_err is not None:
                        das["ye"] = sub_ds[self.y_err]

                    if self.x_err is not None:
                        das["xe"] = sub_ds[self.x_err]

                # nothing to iterate over
                else:
                    das["x"] = self._ds[self.x_coo]
                    das["y"] = self._ds[self.y_coo]

                    if self.c_coo is not None:
                        if mode == "lineplot":
                            self._c_cols.append(
                                np.asscalar(
                                    self._ds[self.c_coo].values.flatten()
                                )
                            )
                        elif mode == "scatter":
                            das["c"] = self._ds[self.c_coo]

                    if self.y_err is not None:
                        das["ye"] = self._ds[self.y_err]

                    if self.x_err is not None:
                        das["xe"] = self._ds[self.x_err]

                for k, da in zip(das, xr.broadcast(*das.values())):
                    data[k] = da.values.flatten()

                # Trim out missing data
                not_null = np.isfinite(data["x"])
                not_null &= np.isfinite(data["y"])

                # TODO: if scatter, broadcast *then* ravel x, y, c?

                data["x"] = data["x"][not_null]
                data["y"] = data["y"][not_null]

                # implement jitter
                if self.xjitter:
                    if self.xlog:
                        data["x"] = data["x"] * np.random.normal(
                            loc=1, scale=self.xjitter, size=data["x"].shape
                        )
                    else:
                        data["x"] = data["x"] + np.random.normal(
                            loc=0, scale=self.xjitter, size=data["x"].shape
                        )

                if self.yjitter:
                    if self.ylog:
                        data["y"] = data["y"] * np.random.normal(
                            loc=1, scale=self.yjitter, size=data["y"].shape
                        )
                    else:
                        data["y"] = data["y"] + np.random.normal(
                            loc=0, scale=self.yjitter, size=data["y"].shape
                        )

                if "c" in data:
                    data["c"] = data["c"][not_null]
                if "ye" in data:
                    data["ye"] = data["ye"][not_null]
                if "xe" in data:
                    data["xe"] = data["xe"][not_null]

                yield data

        self._gen_xy = gen_xy

    def prepare_x_vals_histogram(self):
        """ """

        def gen_x():
            for z in self._z_vals:
                # multiple data variables rather than z coordinate
                if self._multi_var:
                    x = self._ds[z].values.flatten()

                # z-coordinate to iterate over
                elif z is not None:
                    sub_ds = self._ds.loc[{self.z_coo: z}]
                    x = sub_ds[self.x_coo].values.flatten()

                # nothing to iterate over
                else:
                    x = self._ds[self.x_coo].values.flatten()

                yield {"x": x[np.isfinite(x)]}

        self._gen_xy = gen_x

    def prepare_heatmap_data(self, grid=False):
        """Prepare the data to go into a heatmap."""
        # Can only show one dataset
        self._multi_var = False

        self._heatmap_x = self._ds[self.x_coo].values.flatten()
        self._heatmap_y = self._ds[self.y_coo].values.flatten()

        if grid:
            self._zmin = self._ds[self.z_coo].min()
            self._zmax = self._ds[self.z_coo].max()
            return

        self._heatmap_var = ma.masked_invalid(
            self._ds[self.z_coo]
            .squeeze()
            .transpose(self.y_coo, self.x_coo)
            .values
        )
        self._zmin = self._heatmap_var.min()
        self._zmax = self._heatmap_var.max()

    def calc_color_norm(self):
        import matplotlib as mpl

        self.cmap = xyz_colormaps(self.colormap, reverse=self.colormap_reverse)

        coo = self.z_coo if self.c_coo is None else self.c_coo

        if coo is None:
            return

        # check if real numeric type
        if self._ds[coo].dtype.kind in {"i", "u", "f"}:
            self._zmin = self.zlims[0]
            if self._zmin is None:
                self._zmin = self._ds[coo].min().values.item(0)
            self._zmax = self.zlims[1]
            if self._zmax is None:
                self._zmax = self._ds[coo].max().values.item(0)
        else:
            # no relative coloring possible e.g. for strings
            self._zmin, self._zmax = 0.0, 1.0

        if self.vmin is None:
            self.vmin = self._zmin
        if self.vmax is None:
            self.vmax = self._zmax

        self._color_norm = getattr(
            mpl.colors, "LogNorm" if self.colormap_log else "Normalize"
        )(vmin=self.vmin, vmax=self.vmax)

        self.set_mappable()

    def calc_line_colors(self):
        """Helper function for calculating what the colormapped color of each
        line should be.
        """
        self.calc_color_norm()

        # set from 'adjacent' variable
        if self.c_coo is not None:
            rvals = (self._color_norm(z) for z in self._c_cols)
        # set from coordinate
        else:
            if np.isreal(self._z_vals[0]):
                rvals = (self._color_norm(z) for z in self._z_vals)
            else:
                # no relative coloring possible e.g. for strings
                rvals = np.linspace(0, 1, len(self._z_vals))

        # Map relative value to mpl color
        self._cols = (self.cmap(rval) for rval in rvals)
        # Convert colors to correct format
        self._cols = iter(convert_colors(self._cols, outformat=self.backend))

    def prepare_colors(self):
        """Prepare the colors for the lines, based on the z-coordinate."""
        # Automatic colors
        if (self.colors is True) or (self.c_coo is not None):
            self.calc_line_colors()
        # Manually specified colors
        elif self.colors:
            self._cols = itertools.cycle(
                convert_colors(self.colors, outformat=self.backend)
            )
        # Use sequential
        else:
            self._cols = get_default_sequential_cm(self.backend)

    def prepare_markers(self):
        """Prepare the markers to be used for each line."""
        if self.markers is None:
            # If no lines, always draw markers
            if self.lines is False:
                self.markers = True
            # Else decide on how many points
            else:
                self.markers = len(self._ds[self.x_coo]) <= 51

        # Could add more logic based on number of xpoints?
        if self.marker_size is None:
            self._marker_size = 5
        else:
            self._marker_size = self.marker_size

        # decide markers automatically
        if self.markers is True:
            if len(self._z_vals) > 1:
                self._mrkrs = itertools.cycle(_MARKERS[self.backend])
            else:
                self._mrkrs = iter(_SINGLE_LINE_MARKER[self.backend])

        # custom sequence
        elif self.markers:
            self._mrkrs = itertools.cycle(self.markers)

        # no markers
        else:
            self._mrkrs = itertools.repeat(None)

    def prepare_line_styles(self):
        """Line widths and styles."""
        self._lines = (
            itertools.repeat(" ")
            if self.lines is False
            else itertools.repeat("solid")
            if self.line_styles is None
            else itertools.cycle(self.line_styles)
        )

        # Set custom widths for each line
        if self.line_widths is not None:
            self._lws = itertools.cycle(self.line_widths)
        else:
            self._lws = itertools.cycle([1.3])

    def prepare_zorders(self):
        """What order lines appear over one another"""
        if self.zorders is not None:
            self._zordrs = itertools.cycle(self.zorders)
        else:
            self._zordrs = itertools.cycle([3])

    def calc_data_range(self, force=False):
        """ """
        if not self._data_range_calculated or force:
            # x data range
            if self.x_coo is None:
                self._data_xmin, self._data_xmax = None, None
            elif isinstance(self.x_coo, str):
                self._data_xmin = float(self._ds[self.x_coo].min())
                self._data_xmax = float(self._ds[self.x_coo].max())
            else:
                self._data_xmin = float(
                    min(self._ds[x].min() for x in self.x_coo)
                )
                self._data_xmax = float(
                    max(self._ds[x].max() for x in self.x_coo)
                )

            # y data range
            if self.y_coo is None:
                self._data_ymin, self._data_ymax = None, None
            elif isinstance(self.y_coo, str):
                self._data_ymin = float(self._ds[self.y_coo].min())
                self._data_ymax = float(self._ds[self.y_coo].max())
            else:
                self._data_ymin = float(
                    min(self._ds[y].min() for y in self.y_coo)
                )
                self._data_ymax = float(
                    max(self._ds[y].max() for y in self.y_coo)
                )

            self._data_range_calculated = True

    def calc_plot_range(self):
        """Logic for processing limits and padding into plot ranges."""
        # Leave as default
        if self.xlims is None and self.padding is None:
            self._xlims = None
        else:
            if self.xlims is not None:
                xmin, xmax = self.xlims
            else:
                self.calc_data_range()
                xmin, xmax = self._data_xmin, self._data_xmax

            if xmin is None and xmax is None:
                self._xlims = None

            # increase plot range if padding specified
            elif self.padding is not None:
                xrnge = xmax - xmin
                self._xlims = (
                    xmin - self.padding * xrnge,
                    xmax + self.padding * xrnge,
                )
            else:
                self._xlims = xmin, xmax

        if self.ylims is None and self.padding is None:
            # Leave as default
            self._ylims = None
        else:
            if self.ylims is not None:
                ymin, ymax = self.ylims
            else:
                self.calc_data_range()
                ymin, ymax = self._data_ymin, self._data_ymax

            if ymin is None and ymax is None:
                self._ylims = None

            # increase plot range if padding specified
            elif self.padding is not None:
                yrnge = ymax - ymin
                self._ylims = (
                    ymin - self.padding * yrnge,
                    ymax + self.padding * yrnge,
                )
            else:
                self._ylims = ymin, ymax


# Abstract Plotters - contain the core preparation required for each plot type


class AbstractLinePlot:
    def prepare_data_single(self):
        self.prepare_axes_labels()
        self.prepare_z_vals(mode="lineplot")
        self.prepare_z_labels()
        self.calc_use_legend_or_colorbar()
        self.prepare_xy_vals_lineplot(mode="lineplot")
        self.prepare_colors()
        self.prepare_markers()
        self.prepare_line_styles()
        self.prepare_zorders()
        self.calc_plot_range()

    def prepare_data_multi_grid(self):
        self.prepare_axes_labels()
        self.prepare_z_vals(mode="lineplot", grid=True)
        self.calc_use_legend_or_colorbar()
        self.calc_color_norm()
        self.calc_data_range()


class AbstractScatter:
    def prepare_data_single(self):
        self.prepare_axes_labels()
        self.prepare_z_vals(mode="scatter")
        self.prepare_z_labels()
        self.calc_use_legend_or_colorbar()
        self.prepare_xy_vals_lineplot(mode="scatter")
        self.prepare_colors()
        self.prepare_markers()
        self.prepare_line_styles()
        self.prepare_zorders()
        self.calc_plot_range()

    def prepare_data_multi_grid(self):
        self.prepare_axes_labels()
        self.prepare_z_vals(mode="scatter", grid=True)
        self.calc_use_legend_or_colorbar()
        self.calc_color_norm()
        self.calc_data_range()


class AbstractHistogram:
    def prepare_data_single(self):
        self.prepare_axes_labels()
        self.prepare_z_vals(mode="histogram")
        self.prepare_z_labels()
        self.calc_use_legend_or_colorbar()
        self.prepare_x_vals_histogram()
        self.prepare_colors()
        self.prepare_line_styles()
        self.prepare_zorders()
        self.calc_plot_range()

    def prepare_data_multi_grid(self):
        self.prepare_axes_labels()
        self.prepare_z_vals(mode="histogram", grid=True)
        self.calc_use_legend_or_colorbar()
        self.calc_color_norm()


class AbstractHeatMap:
    def prepare_data_single(self):
        self.prepare_axes_labels()
        self.prepare_heatmap_data()
        self.calc_use_legend_or_colorbar()
        self.calc_plot_range()

    def prepare_data_multi_grid(self):
        self.prepare_axes_labels()
        self.prepare_heatmap_data(grid=True)
        self.calc_use_legend_or_colorbar()
        self.calc_color_norm()
        self.calc_data_range()


# Helpers for grid plots


def calc_row_col_datasets(ds, row=None, col=None):
    """ """
    if row is not None:
        rs = ds[row].values
        nr = len(rs)

    if col is not None:
        cs = ds[col].values
        nc = len(cs)

    if row is None:
        return [[ds.loc[{col: c}] for c in cs]], 1, nc

    if col is None:
        return [[ds.loc[{row: r}]] for r in rs], nr, 1

    return [[ds.loc[{row: r, col: c}] for c in cs] for r in rs], nr, nc


def intercept_call_arg(fn):

    @functools.wraps(fn)
    def wrapped_fn(*args, **kwargs):
        call = kwargs.pop("call", True)
        P = fn(*args, **kwargs)

        if call == "both":
            P()
            return P
        elif call:
            return P()
        else:
            return P

    return wrapped_fn


def prettify(x):

    if np.issubdtype(type(x), np.floating):
        x = "{0:0.4f}".format(x).rstrip("0")
        if x[-1] == ".":
            x += "0"

    return x



import collections
import functools
import inspect
import operator
import warnings


@functools.lru_cache(None)
def get_ray():
    """Import and return the ``ray`` module (cached)."""
    import ray

    return ray


class RayFuture:
    """Basic ``concurrent.futures`` like future wrapping a ray ``ObjectRef``."""

    __slots__ = ("_obj", "_cancelled")

    def __init__(self, obj):
        self._obj = obj
        self._cancelled = False

    def result(self, timeout=None):
        return get_ray().get(self._obj, timeout=timeout)

    def done(self):
        return self._cancelled or bool(
            get_ray().wait([self._obj], timeout=0)[0]
        )

    def cancel(self):
        get_ray().cancel(self._obj)
        self._cancelled = True


def _unpack_futures_tuple(x):
    return tuple(map(_unpack_futures, x))


def _unpack_futures_list(x):
    return list(map(_unpack_futures, x))


def _unpack_futures_dict(x):
    return {k: _unpack_futures(v) for k, v in x.items()}


def _unpack_futures_identity(x):
    return x


_unpack_dispatch = collections.defaultdict(
    lambda: _unpack_futures_identity,
    {
        RayFuture: operator.attrgetter("_obj"),
        tuple: _unpack_futures_tuple,
        list: _unpack_futures_list,
        dict: _unpack_futures_dict,
    },
)


def _unpack_futures(x):
    """Allows passing futures by reference - takes e.g. args and kwargs and
    replaces all ``RayFuture`` objects with their underyling ``ObjectRef``
    within all nested tuples, lists and dicts.

    [Subclassing ``ObjectRef`` might avoid needing this.]
    """
    return _unpack_dispatch[x.__class__](x)


@functools.lru_cache(2**14)
def get_remote_fn(fn, **remote_opts):
    """Cached retrieval of remote function."""
    ray = get_ray()
    if remote_opts:
        return ray.remote(**remote_opts)(fn)
    return ray.remote(fn)


@functools.lru_cache(2**14)
def get_fn_as_remote_object(fn):
    """Store ``fn`` in the Ray object store and return an ``ObjectRef``."""
    ray = get_ray()
    return ray.put(fn)


@functools.lru_cache(None)
def get_deploy(**remote_opts):
    """Alternative for 'non-function' callables - e.g. partial
    functions - pass the callable object too.
    """
    ray = get_ray()

    def deploy(fn, *args, **kwargs):
        return fn(*args, **kwargs)

    if remote_opts:
        return ray.remote(**remote_opts)(deploy)
    return ray.remote(deploy)


class RayExecutor:
    """Basic ``concurrent.futures`` like interface using ``ray``.

    Example usage::

        from xyzpy import RayExecutor

        # create a pool that by default requests a single gpu per task
        pool = RayExecutor(
            num_cpus=4,
            num_gpus=4,
            default_remote_opts={"num_gpus": 1},
        )

    """

    def __init__(self, *args, default_remote_opts=None, **kwargs):
        ray = get_ray()
        if not ray.is_initialized():
            ray.init(*args, **kwargs)
        elif args or kwargs:
            warnings.warn(
                "Ignoring arguments to RayExecutor constructor, "
                "ray is already initialized."
            )

        self.default_remote_opts = (
            {} if default_remote_opts is None else dict(default_remote_opts)
        )

    def _maybe_inject_remote_opts(self, remote_opts=None):
        """Return the default remote options, possibly overriding some with
        those supplied by a ``submit call``.
        """
        ropts = self.default_remote_opts
        if remote_opts is not None:
            ropts = {**ropts, **remote_opts}
        return ropts

    def submit(self, fn, *args, pure=False, remote_opts=None, **kwargs):
        """Remotely run ``fn(*args, **kwargs)``, returning a ``RayFuture``."""
        # want to pass futures by reference
        args = _unpack_futures_tuple(args)
        kwargs = _unpack_futures_dict(kwargs)

        ropts = self._maybe_inject_remote_opts(remote_opts)

        # this is the same test ray uses to accept functions
        if inspect.isfunction(fn):
            # can use the faster cached remote function
            obj = get_remote_fn(fn, **ropts).remote(*args, **kwargs)
        else:
            fn_obj = get_fn_as_remote_object(fn)
            obj = get_deploy(**ropts).remote(fn_obj, *args, **kwargs)

        return RayFuture(obj)

    def map(self, func, *iterables, remote_opts=None):
        """Remote map ``func`` over arguments ``iterables``."""
        ropts = self._maybe_inject_remote_opts(remote_opts)
        remote_fn = get_remote_fn(func, **ropts)
        objs = tuple(map(remote_fn.remote, *iterables))
        ray = get_ray()
        return map(ray.get, objs)

    def scatter(self, data):
        """Push ``data`` into the distributed store, returning an ``ObjectRef``
        that can be supplied to ``submit`` calls for example.
        """
        ray = get_ray()
        return ray.put(data)

    def shutdown(self):
        """Shutdown the parent ray cluster, this ``RayExecutor`` instance
        itself does not need any cleanup.
        """
        get_ray().shutdown()


class RayGPUExecutor(RayExecutor):
    """A ``RayExecutor`` that by default requests a single gpu per task."""

    def __init__(self, *args, gpus_per_task=1, **kwargs):
        super().__init__(*args, **kwargs)
        self.default_remote_opts.setdefault("num_gpus", gpus_per_task)



"""Functions for systematically evaluating a function over all combinations."""

import functools
import itertools
import multiprocessing
import random

import numpy as np
import xarray as xr
from joblib.externals.loky import get_reusable_executor

from ..utils import progbar
from .prepare import (
    parse_cases,
    parse_combo_results,
    parse_combos,
    parse_constants,
    parse_resources,
    parse_var_coords,
    parse_var_dims,
    parse_var_names,
)


def infer_shape(x):
    """Take a nested sequence and find its shape as if it were an array.

    Examples
    --------

        >>> x = [[10, 20, 30], [40, 50, 60]]
        >>> infer_shape(x)
        (2, 3)
    """
    shape = ()

    if isinstance(x, str):
        return shape

    try:
        shape += (len(x),)
        return shape + infer_shape(x[0])
    except TypeError:
        return shape


def nan_like_result(res):
    """Take a single result of a function evaluation and calculate the same
    sequence of scalars or arrays but filled entirely with ``nan``.

    Examples
    --------

        >>> res = (True, [[10, 20, 30], [40, 50, 60]], -42.0, 'hello')
        >>> nan_like_result(res)
        (array(nan), array([[nan, nan, nan],
                            [nan, nan, nan]]), array(nan), None)

    """
    if isinstance(res, dict):
        res = xr.Dataset(res)

    if isinstance(res, (xr.Dataset, xr.DataArray)):
        return xr.full_like(res, np.nan, dtype=float)

    if isinstance(res, (bool, str)):
        # - nan gets converted to non-null value of 'nan' (str) -> needs None
        # - by covention turn bool arrays to dtype=object -> needs None
        return None

    try:
        return tuple(np.broadcast_to(np.nan, infer_shape(x)) for x in res)
    except TypeError:
        return np.nan


def _submit(executor, fn, *args, **kwds):
    """Default method for submitting to a executor.

    Parameters
    ----------
    executor : pool-executor like
        A ``multiprocessing.pool`` or an pool executor with API matching either
        ``concurrent.futures``, or an ``ipyparallel`` view.
    fn : callable
        The function to submit.
    args :
        Supplied to ``fn``.
    kwds :
        Supplied to ``fn``.

    Returns
    --------
    future
    """
    if isinstance(executor, multiprocessing.pool.Pool):
        return executor.apply_async(fn, args, kwds)

    elif hasattr(executor, "submit"):
        # concurrent.futures like API
        return executor.submit(fn, *args, **kwds)

    elif hasattr(executor, "apply_async"):
        # ipyparallel like API
        return executor.apply_async(fn, *args, **kwds)

    else:
        raise TypeError(
            "The executor supplied, {}, does not have a ``submit``"
            " or ``apply_async`` method.".format(executor)
        )


def _get_result(future):
    # don't using try-except here, because futures might raise themselves
    if hasattr(future, "result"):
        return future.result()
    if hasattr(future, "get"):
        return future.get()
    raise TypeError("Future does not have a `result` or `get` method.")


def _run_linear_executor(
    executor,
    fn,
    settings,
    verbosity=1,
    desc=None,
):

    progbar_opts = {
        "total": len(settings),
        "disable": verbosity <= 0,
        "desc": desc,
    }

    with progbar(**progbar_opts) as pbar:
        if verbosity >= 2:
            pbar.set_postfix_str("Submitting to executor...", refresh=True)
        futures = [_submit(executor, fn, **kws) for kws in settings]
        results_linear = []
        for kws, future in zip(settings, futures):
            if verbosity >= 2:
                pbar.set_postfix_str(str(kws), refresh=True)
            results_linear.append(_get_result(future))
            pbar.update()
        return results_linear


def _run_linear_sequential(fn, settings, verbosity=1, desc=None):
    results_linear = []

    progbar_opts = {
        "total": len(settings),
        "disable": verbosity <= 0,
        "desc": desc,
    }

    with progbar(**progbar_opts) as pbar:
        for kws in settings:
            if verbosity >= 2:
                pbar.set_postfix_str(str(kws), refresh=True)
            results_linear.append(fn(**kws))
            pbar.update()
        return results_linear


def _unflatten(store, all_combo_values, all_nan=None):
    # non-recursive nested accumulation of results into tuple array
    while all_combo_values:
        # pop out the last arg
        *all_combo_values, last = all_combo_values
        for p in itertools.product(*all_combo_values):
            # for each remaining combination, reduce last arg into tuple
            store[p] = tuple(store.pop(p + (v,), all_nan) for v in last)
    return store.pop(())


def combo_runner_core(
    fn,
    combos,
    constants,
    cases=None,
    split=False,
    flat=False,
    shuffle=False,
    parallel=False,
    num_workers=None,
    executor=None,
    verbosity=1,
    desc=None,
    info=None,
):
    if combos:
        combo_args, combo_values = zip(*combos)
    else:
        combo_args, combo_values = (), ()

    if cases is not None:
        cases = tuple(cases)
        case_args = tuple(cases[0].keys())
        case_values = tuple(tuple(c[a] for a in case_args) for c in cases)
        case_coords = {arg: set() for arg in case_args}
    else:
        # single empty case and everything is in the combos
        case_args = ()
        case_values = ((),)
        case_coords = {}

    if not set(case_args).isdisjoint(combo_args):
        raise ValueError(
            f"Variables can't appear in both ``cases`` and ``combos``, "
            f"currently found combo variables {combo_args} and case variables"
            f"{case_args}."
        )

    # order arguments will be iterated over
    fn_args = case_args + combo_args
    # key location for each case to map into array
    locs = []
    # the actual list of all kwargs supplied to each fn call
    settings = []

    for case_params in case_values:
        # keep track of every case value we see to form union later
        for arg, v in zip(case_args, case_params):
            case_coords[arg].add(v)

        for combo_params in itertools.product(*combo_values):
            loc = case_params + combo_params
            kws = dict(zip(fn_args, loc))
            kws.update(constants)
            locs.append(loc)
            settings.append(kws)

    if shuffle:
        random.seed(int(shuffle))
        enum_settings = list(enumerate(settings))
        random.shuffle(enum_settings)
        enum, settings = zip(*enum_settings)

    run_linear_opts = {
        "fn": fn,
        "settings": settings,
        "verbosity": verbosity,
        "desc": desc,
    }

    if executor == "ray":
        from .ray_executor import RayExecutor

        executor = RayExecutor(num_cpus=num_workers)

    if executor is not None:
        # custom pool supplied
        results_linear = _run_linear_executor(executor, **run_linear_opts)
    elif parallel or num_workers:
        # else for parallel, by default use a process pool-exceutor

        if (
            # bools are ints, so check for that first since True != 1 here
            (not isinstance(parallel, bool))
            and isinstance(parallel, int)
            and (num_workers is None)
        ):
            # assume parallel is the number of workers
            num_workers = parallel

        executor = get_reusable_executor(num_workers)
        results_linear = _run_linear_executor(executor, **run_linear_opts)
    else:
        results_linear = _run_linear_sequential(**run_linear_opts)

    if shuffle:
        enum_results = sorted(zip(enum, results_linear), key=lambda x: x[0])
        _, results_linear = zip(*enum_results)

    # try and put the union of case coordinates into a reasonable order
    for arg in case_args:
        try:
            case_coords[arg] = sorted(case_coords[arg])
        except TypeError:  # unsortable
            case_coords[arg] = list(case_coords[arg])

    # find the equivalent combos as if all coordinates had been run
    combos_cases = tuple(case_coords.values())
    all_combo_values = combos_cases + combo_values

    def process_results(r):
        if flat:
            # just return the list of results
            return tuple(r)

        results_mapped = dict(zip(locs, r))

        if cases is None:
            # we ran all combinations -> no missing data
            return _unflatten(results_mapped, combo_values)

        # unpack dict into nested tuple, ready for numpy
        all_nan = nan_like_result(r[0])
        results = _unflatten(results_mapped, all_combo_values, all_nan)

        return results

    if info is not None:
        # optionally return some extra labelling information
        if flat:
            info["settings"] = settings
        else:
            info["fn_args"] = fn_args
            info["all_combo_values"] = all_combo_values

    if split:
        # put each output variable into a seperate results at the top level
        return tuple(process_results(r) for r in zip(*results_linear))
    else:
        return process_results(results_linear)


def combo_runner(
    fn,
    combos=None,
    *,
    cases=None,
    constants=None,
    split=False,
    flat=False,
    shuffle=False,
    parallel=False,
    executor=None,
    num_workers=None,
    verbosity=1,
    desc=None,
):
    """Take a function ``fn`` and compute it over all combinations of named
    variables values, optionally showing progress and in parallel.

    Parameters
    ----------
    fn : callable
        Function to analyse.
    combos : dict_like[str, iterable]
        All combinations of each argument to values mapping will be computed.
        Each argument range thus gets a dimension in the output array(s).
    cases  : sequence of mappings, optional
        Optional list of specific configurations. If both ``combos`` and
        ``cases`` are given, then the function is computed for all
        sub-combinations in ``combos`` for each case in ``cases``, arguments
        can thus only appear in one or the other. Note that missing
        combinations of arguments will be represented by ``nan`` if creating a
        nested array.
    constants : dict, optional
        Constant function arguments. Unlike ``combos`` and ``cases``, these
        won't produce dimensions in the output result when ``flat=False``.
    split : bool, optional
        Whether to split (unzip) the outputs of ``fn`` into multiple output
        arrays or not.
    flat : bool, optional
        Whether to return a flat list of results or to return a nested
        tuple suitable to be supplied to ``numpy.array``.
    shuffle : bool or int, optional
        If given, compute the results in a random order (using ``random.seed``
        and ``random.shuffle``), which can be helpful for distributing
        resources when not all cases are computationally equal.
    parallel : bool, optional
        Process combos in parallel, default number of workers picked.
    executor : executor-like pool, optional
        Submit all combos to this pool executor. Must have ``submit`` or
        ``apply_async`` methods and API matching either ``concurrent.futures``
        or an ``ipyparallel`` view. Pools from ``multiprocessing.pool`` are
        also  supported.
    num_workers : int, optional
        Explicitly choose how many workers to use, None for automatic.
    verbosity : {0, 1, 2}, optional
        How much information to display:

        - 0: nothing,
        - 1: just progress,
        - 2: postfix the current settings to the progress bar.

    desc : str, optional
        Description to show in the progress bar, if ``verbosity > 0``.

    Returns
    -------
    data : nested tuple
        Nested tuple containing all combinations of running ``fn`` if
        ``flat == False`` else a flat list of results.

    Examples
    --------

        >>> def fn(a, b, c, d):
        ...     return str(a) + str(b) + str(c) + str(d)


    Run all possible combos::

        >>> xyz.combo_runner(
        ...     fn,
        ...     combos={
        ...         'a': [1, 2],
        ...         'b': [3, 4],
        ...         'c': [5, 6],
        ...         'd': [7, 8],
        ...     },
        ... )
        100%|##########| 16/16 [00:00<00:00, 84733.41it/s]

        (((('1357', '1358'), ('1367', '1368')),
          (('1457', '1458'), ('1467', '1468'))),
         ((('2357', '2358'), ('2367', '2368')),
          (('2457', '2458'), ('2467', '2468'))))

    Run only a selection of cases::

        >>> xyz.combo_runner(
        ...     fn,
        ...     cases=[
        ...         {'a': 1, 'b': 3, 'c': 5, 'd': 7},
        ...         {'a': 2, 'b': 4, 'c': 6, 'd': 8},
        ...     ],
        ... )
        100%|##########| 2/2 [00:00<00:00, 31418.01it/s]
        (((('1357', nan), (nan, nan)),
          ((nan, nan), (nan, nan))),
         (((nan, nan), (nan, nan)),
          ((nan, nan), (nan, '2468'))))

    Run only certain cases of some args, but all combinations of others::

        >>> xyz.combo_runner(
        ...     fn,
        ...     cases=[
        ...         {'a': 1, 'b': 3},
        ...         {'a': 2, 'b': 4},
        ...     ],
        ...     combos={
        ...         'c': [3, 4],
        ...         'd': [4, 5],
        ...     },
        ... )
        100%|##########| 8/8 [00:00<00:00, 92691.80it/s]
        (((('1334', '1335'), ('1344', '1345')),
          ((nan, nan), (nan, nan))),
         (((nan, nan), (nan, nan)),
          (('2434', '2435'), ('2444', '2445'))))

    """
    # Prepare combos
    cases = parse_cases(cases)
    combos = parse_combos(combos)
    constants = parse_constants(constants)

    # Submit to core combo runner
    return combo_runner_core(
        fn=fn,
        combos=combos,
        cases=cases,
        constants=constants,
        split=split,
        flat=flat,
        shuffle=shuffle,
        parallel=parallel,
        executor=executor,
        num_workers=num_workers,
        verbosity=verbosity,
        desc=desc,
    )


def multi_concat(results, dims):
    """Concatenate a nested list of xarray objects along several dimensions."""
    if len(dims) == 1:
        return xr.concat(
            [
                # if a dict, convert to dataset
                xr.Dataset(obj)
                if isinstance(obj, dict)
                # else assume it's a dataset or datarray
                else obj
                for obj in results
            ],
            dim=dims[0],
        )
    else:
        return xr.concat(
            [multi_concat(sub_results, dims[1:]) for sub_results in results],
            dim=dims[0],
        )


def get_ndim_first(x, ndim):
    """Return the first element from the ndim-nested list x."""
    return x if ndim == 0 else get_ndim_first(x[0], ndim - 1)


def results_to_ds(
    results,
    combos,
    var_names,
    var_dims,
    var_coords,
    constants=None,
    attrs=None,
):
    """Convert the output of combo_runner into a :class:`xarray.Dataset`."""
    fn_args = tuple(x for x, _ in combos)
    results = parse_combo_results(results, var_names)

    if len(results) != len(var_names):
        raise ValueError(
            f"Wrong number of results ({len(results)}) for "
            f"{len(var_names)} ``var_names``: {var_names}."
        )

    # Check if the results are an array of xarray objects
    xobj_results = isinstance(
        get_ndim_first(results, len(fn_args) + 1),
        (dict, xr.Dataset, xr.DataArray),
    )

    if xobj_results:
        # concat them all together, no var_names needed
        ds = multi_concat(results[0], fn_args)
        # Set dataset coordinates
        for fn_arg, vals in combos:
            ds[fn_arg] = vals
    else:
        # create a new dataset using the given arrays and var_names
        ds = xr.Dataset(
            coords={**dict(combos), **dict(var_coords)},
            data_vars={
                name: (fn_args + var_dims[name], np.asarray(data))
                for data, name in zip(results, var_names)
            },
        )

    if attrs:
        ds.attrs = attrs

    # Add constants to attrs, but filter out those which should be coords
    if constants:
        for k, v in constants.items():
            if callable(v):
                # should't store functions as attrs or coords
                continue

            if k in ds.dims:
                ds.coords[k] = v
            else:
                try:
                    ds.attrs[k] = v
                except Exception as e:
                    import warnings

                    warnings.warn(
                        f"Failed to add constant {k}={v} to dataset attrs: {e}"
                    )
    return ds


def results_to_df(
    results_linear,
    settings,
    attrs,
    resources,
    var_names,
):
    """Convert the output of combo_runner into a :class:`pandas.DataFrame`."""
    import pandas as pd

    # construct as list of single dict entries
    data = []
    for row, result in zip(settings, results_linear):
        # don't record resources
        for k in resources:
            row.pop(k, None)

        # add in the attrs, note this isn't quite equivalent to dataset case,
        # as we add the attributes for every entry -> limitation of dataframe
        if attrs:
            row.update(attrs)

        # add in the output variables
        try:
            row.update(dict(zip(var_names, result)))
        except TypeError:
            row.update(dict(zip(var_names, [result])))

        data.append(row)

    # convert to dataframe
    return pd.DataFrame(data)


def combo_runner_to_ds(
    fn,
    combos,
    var_names,
    *,
    var_dims=None,
    var_coords=None,
    cases=None,
    constants=None,
    resources=None,
    attrs=None,
    shuffle=False,
    parse=True,
    to_df=False,
    parallel=False,
    num_workers=None,
    executor=None,
    verbosity=1,
    desc=None,
):
    """Evaluate a function over all cases and combinations and output to a
    :class:`xarray.Dataset`.

    Parameters
    ----------
    fn : callable
        Function to evaluate.
    combos : dict_like[str, iterable]
        Mapping of each individual function argument to sequence of values.
    var_names : str, sequence of strings, or None
        Variable name(s) of the output(s) of `fn`, set to None if
        fn outputs data already labelled in a Dataset or DataArray.
    var_dims : sequence of either strings or string sequences, optional
        'Internal' names of dimensions for each variable, the values for
        each dimension should be contained as a mapping in either
        `var_coords` (not needed by `fn`) or `constants` (needed by `fn`).
    var_coords : mapping, optional
        Mapping of extra coords the output variables may depend on.
    cases : sequence of dicts, optional
        Individual cases to run for some or all function arguments.
    constants : mapping, optional
        Arguments to `fn` which are not iterated over, these will be
        recorded either as attributes or coordinates if they are named
        in `var_dims`.
    resources : mapping, optional
        Like `constants` but they will not be recorded.
    attrs : mapping, optional
        Any extra attributes to store.
    parallel : bool, optional
        Process combos in parallel, default number of workers picked.
    executor : executor-like pool, optional
        Submit all combos to this pool executor. Must have ``submit`` or
        ``apply_async`` methods and API matching either ``concurrent.futures``
        or an ``ipyparallel`` view. Pools from ``multiprocessing.pool`` are
        also  supported.
    num_workers : int, optional
        Explicitly choose how many workers to use, None for automatic.
    verbosity : {0, 1, 2}, optional
        How much information to display:

        - 0: nothing,
        - 1: just progress,
        - 2: postfix the current settings to the progress bar.

    desc : str, optional
        Description to show in the progress bar, if ``verbosity > 0``.

    Returns
    -------
    ds : xarray.Dataset or pandas.DataFrame
        Multidimensional labelled dataset contatining all the results if
        ``to_df=False`` (the default), else a pandas dataframe with results
        as labelled rows.
    """
    if to_df:
        if var_names is None:
            raise ValueError("Can't coerce dataset output into dataframe.")
        if var_dims is not None and any(var_dims.values()):
            raise ValueError("Dataframes don't support internal dimensions.")
        if var_coords:
            raise ValueError("Dataframes don't support internal dimensions.")

    if parse:
        combos = parse_combos(combos)
        cases = parse_cases(cases)
        constants = parse_constants(constants)
        resources = parse_resources(resources)
        var_names = parse_var_names(var_names)
        var_dims = parse_var_dims(var_dims, var_names=var_names)
        var_coords = parse_var_coords(var_coords)

    if (cases is not None) or to_df:
        info = {}
    else:
        info = None

    # Generate data for all combos
    results = combo_runner_core(
        fn=fn,
        combos=combos,
        cases=cases,
        constants={**resources, **constants},
        parallel=parallel,
        num_workers=num_workers,
        executor=executor,
        verbosity=verbosity,
        desc=desc,
        info=info,
        split=(not to_df) and (len(var_names) > 1),
        flat=to_df,
        shuffle=shuffle,
    )

    if to_df:
        # convert flat tuple of results to dataframe
        return results_to_df(
            results,
            settings=info["settings"],
            attrs=attrs,
            resources=resources,
            var_names=var_names,
        )

    if cases is not None:
        # if we have cases, then need to find the effective full combos
        # -> results contains nan placeholders for non-run cases
        combos = tuple(zip(info["fn_args"], info["all_combo_values"]))

    # convert to dataset
    ds = results_to_ds(
        results,
        combos,
        var_names=var_names,
        var_dims=var_dims,
        var_coords=var_coords,
        constants=constants,
        attrs=attrs,
    )

    return ds


combo_runner_to_df = functools.partial(combo_runner_to_ds, to_df=True)



"""Utility functions."""

import functools
import itertools
import operator
import sys
import time
from collections.abc import Iterable

import numpy as np
import tqdm


class XYZError(Exception):
    pass


def isiterable(obj):
    return isinstance(obj, Iterable)


def prod(it):
    """Product of an iterable."""
    return functools.reduce(operator.mul, it)


def unzip(its, zip_level=1):
    """Split a nested iterable at a specified level, i.e. in numpy language
    transpose the specified 'axis' to be the first.

    Parameters
    ----------
    its: iterable (of iterables (of iterables ...))
        'n-dimensional' iterable to split
    zip_level: int
        level at which to split the iterable, default of 1 replicates
        ``zip(*its)`` behaviour.

    Example
    -------
    >>> x = [[(1, True), (2, False), (3, True)],
             [(7, True), (8, False), (9, True)]]
    >>> nums, bools = unzip(x, 2)
    >>> nums
    ((1, 2, 3), (7, 8, 9))
    >>> bools
    ((True, False, True), (True, False, True))

    """

    def _unzipper(its, zip_level):
        if zip_level > 1:
            return (zip(*_unzipper(it, zip_level - 1)) for it in its)
        else:
            return its

    return zip(*_unzipper(its, zip_level)) if zip_level else its


def flatten(its, n):
    """Take the n-dimensional nested iterable its and flatten it.

    Parameters
    ----------
        its : nested iterable
        n : number of dimensions

    Returns
    -------
        flattened iterable of all items
    """
    if n > 1:
        return itertools.chain(*(flatten(it, n - 1) for it in its))
    else:
        return its


def _get_fn_name(fn):
    """Try to inspect a function's name, taking into account several common
    non-standard types of function: dask, functools.partial ...
    """
    if hasattr(fn, "__name__"):
        return fn.__name__
    # try dask delayed function with key
    elif hasattr(fn, "key"):
        return fn.key.partition("-")[0]
    # try functools.partial function syntax
    elif hasattr(fn, "func"):
        return fn.func.__name__
    else:
        raise ValueError("Could not extract function name from {}".format(fn))


def progbar(it=None, nb=False, **kwargs):
    """Turn any iterable into a progress bar, with notebook option

    Parameters
    ----------
        it: iterable
            Iterable to wrap with progress bar
        nb: bool
            Whether  to display the notebook progress bar
        **kwargs: dict-like
            additional options to send to tqdm
    """
    defaults = {"ascii": True, "smoothing": 0.0}
    # Overide defaults with custom kwargs
    settings = {**defaults, **kwargs}
    if nb:  # pragma: no cover
        return tqdm.tqdm_notebook(it, **settings)
    return tqdm.tqdm(it, **settings)


def getsizeof(obj):
    """Compute the real size of a Python object in bytes, taken from
    https://stackoverflow.com/a/30316760/5640201.

    Parameters
    ----------
    obj : object
        Object to measure.

    Returns
    -------
    int
        Total size in bytes.
    """
    import sys
    from gc import get_referents
    from types import FunctionType, ModuleType

    # Custom objects know their class.
    # Function objects seem to know way too much, including modules.
    # Exclude modules as well.
    excluded = type, ModuleType, FunctionType
    if isinstance(obj, excluded):
        raise TypeError(
            "getsize() does not take argument of type: {}".format(type(obj))
        )

    seen_ids = set()
    size = 0
    objects = [obj]
    while objects:
        need_referents = []
        for obj in objects:
            if not isinstance(obj, excluded) and id(obj) not in seen_ids:
                seen_ids.add(id(obj))
                size += sys.getsizeof(obj)
                need_referents.append(obj)
        objects = get_referents(*need_referents)
    return size


class Timer:
    """A very simple context manager class for timing blocks.

    Examples
    --------

    >>> from xyzpy import Timer
    >>> with Timer() as timer:
    ...     print('Doing some work!')
    ...
    Doing some work!
    >>> timer.t
    0.00010752677917480469

    """

    def __enter__(self):
        self.start = time.time()
        return self

    def __exit__(self, *args):
        self.end = time.time()
        self.t = self.time = self.interval = self.end - self.start


def _auto_min_time(timer, min_t=0.2, repeats=2, get="min"):
    tot_t = 0
    number = 1

    while True:
        tot_t = timer.timeit(number)
        if tot_t > min_t:
            break
        number *= 2

    results = [tot_t] + timer.repeat(repeats - 1, number)

    if get == "mean":
        return sum(results) / (number * len(results))

    return min(t / number for t in results)


def benchmark(
    fn, setup=None, n=None, min_t=0.2, repeats=2, get="min", starmap=False
):
    """Benchmark the time it takes to run ``fn``.

    Parameters
    ----------
    fn : callable
        The function to time.
    setup : callable, optional
        If supplied the function that sets up the argument for ``fn``.
    n : int, optional
        If supplied, the integer to supply to ``setup`` of ``fn``.
    min_t : float, optional
        Aim to repeat function enough times to take up this many seconds.
    repeats : int, optional
        Repeat the whole procedure (with setup) this many times in order to
        take the minimum run time.
    get : {'min', 'mean'}, optional
        Return the minimum or mean time for each run.
    starmap : bool, optional
        Unpack the arguments from ``setup``, if given.

    Returns
    -------
    t : float
        The minimum, averaged, time to run ``fn`` in seconds.

    Examples
    --------

    Just a parameter-less function:

        >>> import xyzpy as xyz
        >>> import numpy as np
        >>> xyz.benchmark(lambda: np.linalg.eig(np.random.randn(100, 100)))
        0.004726233000837965

    The same but with a setup and size parameter ``n`` specified:

        >>> setup = lambda n: np.random.randn(n, n)
        >>> fn = lambda X: np.linalg.eig(X)
        >>> xyz.benchmark(fn, setup, 100)
        0.0042192734545096755
    """
    from timeit import Timer

    if n is None:
        n = ""

    if setup is None:
        setup_str = ""
        stmnt_str = "fn({})".format(n)
    else:
        setup_str = "X=setup({})".format(n)
        stmnt_str = "fn(*X)" if starmap else "fn(X)"

    timer = Timer(
        setup=setup_str, stmt=stmnt_str, globals={"setup": setup, "fn": fn}
    )

    return _auto_min_time(timer, min_t=min_t, repeats=repeats, get=get)


class Benchmarker:
    """Compare the performance of various ``kernels``. Internally this makes
    use of :func:`~xyzpy.benchmark`, :func:`~xyzpy.Harvester` and xyzpys
    plotting functionality.

    Parameters
    ----------
    kernels : sequence of callable
        The functions to compare performance with.
    setup : callable, optional
        If given, setup each benchmark run by suppling the size argument ``n``
        to this function first, then feeding its output to each of the
        functions.
    names : sequence of str, optional
        Alternate names to give the function, else they will be inferred.
    benchmark_opts : dict, optional
        Supplied to :func:`~xyzpy.benchmark`.
    data_name : str, optional
        If given, the file name the internal harvester will use to store
        results persistently.

    Attributes
    ----------
    harvester : xyz.Harvester
        The harvester that runs and accumulates all the data.
    ds : xarray.Dataset
        Shortcut to the harvester's full dataset.
    """

    def __init__(
        self,
        kernels,
        setup=None,
        names=None,
        benchmark_opts=None,
        data_name=None,
    ):
        import xyzpy as xyz

        self.kernels = kernels
        self.names = [f.__name__ for f in kernels] if names is None else names
        self.setup = setup
        self.benchmark_opts = {} if benchmark_opts is None else benchmark_opts

        def time(n, kernel):
            fn = self.kernels[self.names.index(kernel)]
            return xyz.benchmark(fn, self.setup, n, **self.benchmark_opts)

        self.runner = xyz.Runner(time, ["time"])
        self.harvester = xyz.Harvester(self.runner, data_name=data_name)

    def run(self, ns, kernels=None, **harvest_opts):
        """Run the benchmarks. Each run accumulates rather than overwriting the
        results.

        Parameters
        ----------
        ns : sequence of int or int
            The sizes to run the benchmarks with.
        kernels : sequence of str, optional
            If given, only run the kernels with these names.
        harvest_opts
            Supplied to :meth:`~xyzpy.Harvester.harvest_combos`.
        """
        if not isiterable(ns):
            ns = (ns,)

        if kernels is None:
            kernels = self.names

        combos = {"n": ns, "kernel": kernels}
        self.harvester.harvest_combos(combos, **harvest_opts)

    @property
    def ds(self):
        return self.harvester.full_ds

    def plot(self, **plot_opts):
        """Plot the benchmarking results."""
        plot_opts.setdefault("xscale", "log")
        plot_opts.setdefault("yscale", "log")
        return self.ds.xyz.plot(x="n", y="time", color="kernel", **plot_opts)

    def lineplot(self, **plot_opts):
        """Plot the benchmarking results."""
        plot_opts.setdefault("xlog", True)
        plot_opts.setdefault("ylog", True)
        return self.ds.xyz.lineplot("n", "time", "kernel", **plot_opts)

    def ilineplot(self, **plot_opts):
        """Interactively plot the benchmarking results."""
        plot_opts.setdefault("xlog", True)
        plot_opts.setdefault("ylog", True)
        return self.ds.xyz.ilineplot("n", "time", "kernel", **plot_opts)


def format_number_with_error(x, err):
    """Given ``x`` with error ``err``, format a string showing the relevant
    digits of ``x`` with two significant digits of the error bracketed, and
    overall exponent if necessary.

    Parameters
    ----------
    x : float
        The value to print.
    err : float
        The error on ``x``.

    Returns
    -------
    str

    Examples
    --------

        >>> print_number_with_uncertainty(0.1542412, 0.0626653)
        '0.154(63)'

        >>> print_number_with_uncertainty(-128124123097, 6424)
        '-1.281241231(64)e+11'

    """
    # compute an overall scaling for both values
    x_exponent = max(
        int(f"{x:e}".split("e")[1]),
        int(f"{err:e}".split("e")[1]) + 1,
    )
    # for readability try and show values close to 1 with no exponent
    hide_exponent = (
        # nicer showing 0.xxx(yy) than x.xx(yy)e-1
        (x_exponent in (0, -1))
        or
        # also nicer showing xx.xx(yy) than x.xxx(yy)e+1
        ((x_exponent == +1) and (err < abs(x / 10)))
    )
    if hide_exponent:
        suffix = ""
    else:
        x = x / 10**x_exponent
        err = err / 10**x_exponent
        suffix = f"e{x_exponent:+03d}"

    # work out how many digits to print
    # format the main number and bracketed error
    mantissa, exponent = f"{err:.1e}".split("e")
    mantissa, exponent = mantissa.replace(".", ""), int(exponent)
    return f"{x:.{abs(exponent) + 1}f}({mantissa}){suffix}"


class RunningStatistics:  # pragma: no cover
    """Running mean & standard deviation using Welford's
    algorithm. This is a very efficient way of keeping track of the error on
    the mean for example.

    Attributes
    ----------
    mean : float
        Current mean.
    count : int
        Current count.
    std : float
        Current standard deviation.
    var : float
        Current variance.
    err : float
        Current error on the mean.
    rel_err: float
        The current relative error.

    Examples
    --------

        >>> rs = RunningStatistics()
        >>> rs.update(1.1)
        >>> rs.update(1.4)
        >>> rs.update(1.2)
        >>> rs.update_from_it([1.5, 1.3, 1.6])
        >>> rs.mean
        1.3499999046325684

        >>> rs.std  # standard deviation
        0.17078252585383266

        >>> rs.err  # error on the mean
        0.06972167422092768

    """

    def __init__(self):
        self.count = 0
        self.mean = 0.0
        self.M2 = 0.0

    def update(self, x):
        """Add a single value ``x`` to the statistics."""
        self.count += 1
        delta = x - self.mean
        self.mean += delta / self.count
        delta2 = x - self.mean
        self.M2 += delta * delta2

    def update_from_it(self, xs):
        """Add all values from iterable ``xs`` to the statistics."""
        for x in xs:
            self.update(x)

    def converged(self, rtol, atol):
        """Check if the stats have converged with respect to relative and
        absolute tolerance ``rtol`` and ``atol``.
        """
        return self.err < rtol * abs(self.mean) + atol

    @property
    def var(self):
        if self.count == 0:
            return np.inf
        return self.M2 / self.count

    @property
    def std(self):
        if self.count == 0:
            return np.inf
        return self.var**0.5

    @property
    def err(self):
        if self.count == 0:
            return np.inf
        return self.std / self.count**0.5

    @property
    def rel_err(self):
        if self.count == 0:
            return np.inf
        return self.err / abs(self.mean)

    def __repr__(self):
        if self.count == 0:
            # mean and error are undefined
            return "RunningStatistics(mean=None, count=0)"

        return (
            f"RunningStatistics("
            f"mean={format_number_with_error(self.mean, self.err)}, "
            f"count={self.count}"
            ")"
        )


class RunningCovariance:  # pragma: no cover
    """Running covariance class."""

    def __init__(self):
        self.count = 0
        self.xmean = 0.0
        self.ymean = 0.0
        self.C = 0.0

    def update(self, x, y):
        self.count += 1
        dx = x - self.xmean
        dy = y - self.ymean
        self.xmean += dx / self.count
        self.ymean += dy / self.count
        self.C += dx * (y - self.ymean)

    def update_from_it(self, xs, ys):
        for x, y in zip(xs, ys):
            self.update(x, y)

    @property
    def covar(self):
        """The covariance."""
        return self.C / self.count

    @property
    def sample_covar(self):
        """The covariance with "Bessel's correction"."""
        return self.C / (self.count - 1)


class RunningCovarianceMatrix:
    """Running covariance matrix for ``n`` variables.

    Parameters
    ----------
    n : int, optional
        Number of variables to track.
    """

    def __init__(self, n=2):
        """Initialize the running covariance matrix."""
        self.n = n
        self.rcs = {}
        for i in range(self.n):
            for j in range(i, self.n):
                self.rcs[i, j] = RunningCovariance()

    def update(self, *x):
        """Update the covariance matrix with a single observation."""
        for i in range(self.n):
            for j in range(i, self.n):
                self.rcs[i, j].update(x[i], x[j])

    def update_from_it(self, *xs):
        """Update from iterables of observations for each variable."""
        for i in range(self.n):
            for j in range(i, self.n):
                self.rcs[i, j].update_from_it(xs[i], xs[j])

    @property
    def count(self):
        """Return the number of samples accumulated."""
        return self.rcs[0, 0].count

    @property
    def covar_matrix(self):
        """Return the population covariance matrix."""
        covar_matrix = np.empty((self.n, self.n))
        for i in range(self.n):
            for j in range(self.n):
                if j >= i:
                    covar_matrix[i, j] = self.rcs[i, j].covar
                else:
                    covar_matrix[i, j] = self.rcs[j, i].covar
        return covar_matrix

    @property
    def sample_covar_matrix(self):
        """Return the sample covariance matrix."""
        covar_matrix = np.empty((self.n, self.n))
        for i in range(self.n):
            for j in range(self.n):
                if j >= i:
                    covar_matrix[i, j] = self.rcs[i, j].sample_covar
                else:
                    covar_matrix[i, j] = self.rcs[j, i].sample_covar
        return covar_matrix

    def to_uncertainties(self, bias=True):
        """Convert the accumulated statistics to correlated uncertainties,
        from which new quantities can be calculated with error automatically
        propagated.

        Parameters
        ----------
        bias : bool, optional
            If False, use the sample covariance with "Bessel's correction".

        Return
        ------
        values : tuple of uncertainties.ufloat
            The sequence of correlated variables.

        Examples
        --------

        Estimate quantities of two perfectly correlated sequences.

            >>> rcm = xyz.RunningCovarianceMatrix()
            >>> rcm.update_from_it((1, 3, 2), (2, 6, 4))
            >>> x, y = rcm.to_uncertainties(rcm)

        Calculated quantities like sums have the error propagated:

            >>> x + y
            6.0+/-2.4494897427831783

        But the covariance is also taken into account, meaning the ratio here
        can be estimated with zero error:

            >>> x / y
            0.5+/-0

        """
        import uncertainties

        means = [self.rcs[i, i].xmean for i in range(self.n)]
        if bias:
            covar = self.covar_matrix
        else:
            covar = self.sample_covar_matrix

        return uncertainties.correlated_values(means, covar)


def estimate_from_repeats(
    fn,
    *fn_args,
    rtol=0.02,
    tol_scale=1.0,
    get="stats",
    verbosity=0,
    min_samples=5,
    max_samples=1000000,
    **fn_kwargs,
):
    """
    Parameters
    ----------
    fn : callable
        The function that estimates a single value.
    fn_args, optional
        Supplied to ``fn``.
    rtol : float, optional
        Relative tolerance for error on mean.
    tol_scale : float, optional
        The expected 'scale' of the estimate, this modifies the aboslute
        tolerance near zero to ``rtol * tol_scale``, default: 1.0.
    get : {'stats', 'samples', 'mean'}, optional
        Just get the ``RunningStatistics`` object, or the actual samples too,
        or just the actual mean estimate.
    verbosity : { 0, 1, 2}, optional
        How much information to show:

        - ``0``: nothing
        - ``1``: progress bar just with iteration rate,
        - ``2``: progress bar with running stats displayed.

    min_samples : int, optional
        Take at least this many samples before checking for convergence.
    max_samples : int, optional
        Take at maximum this many samples.
    fn_kwargs, optional
        Supplied to ``fn``.

    Returns
    -------
    rs : RunningStatistics
        Statistics about the random estimation.
    samples : list[float]
        If ``get=='samples'``, the actual samples.


    Examples
    --------

    Estimate the sum of ``n`` random numbers:

        >>> import numpy as np
        >>> import xyzpy as xyz
        >>> def fn(n):
        ...     return np.random.rand(n).sum()
        ...
        >>> stats = xyz.estimate_from_repeats(fn, n=10, verbosity=3)
        59: 5.13(12): : 58it [00:00, 3610.84it/s]
        RunningStatistics(mean=5.13(12), count=59)

    """

    rs = RunningStatistics()
    repeats = itertools.count()

    if verbosity >= 1:
        repeats = progbar(repeats)

    if get == "samples":
        xs = []

    try:
        for i in repeats:
            x = fn(*fn_args, **fn_kwargs)
            if get == "samples":
                xs.append(x)
            rs.update(x)

            if verbosity >= 2:
                repeats.set_description(
                    f"{rs.count}: {format_number_with_error(rs.mean, rs.err)}"
                )

            # need at least min_samples to check convergence
            if i > min_samples:
                if rs.converged(rtol, tol_scale * rtol):
                    break

            # reached the maximum number of samples to try
            if i >= max_samples - 1:
                break
    except KeyboardInterrupt:
        # allow user to cleanly interupt sampling with keyboard
        pass
    finally:
        if verbosity >= 1:
            repeats.close()

    if verbosity >= 1:
        sys.stderr.flush()
        print(rs)

    if get == "samples":
        return rs, xs

    if get == "mean":
        return rs.mean

    return rs


class MemoryMonitor:
    """Monitor this process' peak memory usage with specified sampling interval
    in a daemon thread. This is intended to be used as a context manager for
    long running and memory intensive processes, not fine grained memory
    tracking.

    Parameters
    ----------
    interval : float, optional
        Time between memory measurements in seconds. Fluctuations in peak
        memory between measurements might not be captured.

    Attributes
    ----------
    interval : float
        Time between memory measurements in seconds.
    peak : float
        The peak memory usage in gigabytes.
    """

    def __init__(self, interval: float = 0.1):
        self.interval = interval
        self.peak = None
        self.is_running = False
        self.monitor_thread = None

    def _monitor(self):
        import psutil

        if self.peak is None:
            self.peak = 0.0

        process = psutil.Process()
        while self.is_running:
            current = process.memory_info().rss / (1 << 30)
            self.peak = max(self.peak, current)
            time.sleep(self.interval)

    def start(self):
        """Start the memory monitoring thread."""
        import threading

        self.is_running = True
        self.monitor_thread = threading.Thread(target=self._monitor)
        self.monitor_thread.daemon = True  # don't prevent program exit
        self.monitor_thread.start()

    def stop(self):
        """Stop the memory monitoring thread."""
        self.is_running = False
        if self.monitor_thread:
            self.monitor_thread.join()

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.stop()

    def __del__(self):
        self.stop()

    def __repr__(self):
        s = f"MemoryMonitor(interval={self.interval}, peak={self.peak:.2f}GB)"
        return s


def get_peak_memory_usage():
    """Get the peak memory usage of the current process in *gigabytes*. This
    uses the `psutil` package on Windows, and the `resource` package on
    Linux and macOS.
    """
    if sys.platform == "win32":
        import psutil

        process = psutil.Process()
        mem_info = process.memory_info()
        peak = mem_info.peak_wset / (1 << 30)

    elif sys.platform == "darwin":
        import resource

        usage = resource.getrusage(resource.RUSAGE_SELF)
        # returned in bytes
        peak = usage.ru_maxrss / (1 << 30)

    elif sys.platform.startswith("linux"):
        import resource

        usage = resource.getrusage(resource.RUSAGE_SELF)
        # returned in kilobytes
        peak = usage.ru_maxrss / (1 << 20)

    return peak


def report_memory():
    """Return a formatted memory usage summary for the current process."""
    try:
        import psutil

        # Get total memory
        total_memory = psutil.virtual_memory().total

        # Get memory used
        used_memory = psutil.virtual_memory().used

        # Get memory used by current process
        process_memory = psutil.Process().memory_info().rss

        # Return memory report
        return (
            f"Process memory: {process_memory / 1e9:>10.2f}GB / "
            f"Memory used: {used_memory / 1e9:>10.2f}GB / "
            f"Total memory: {total_memory / 1e9:>10.2f}GB "
        )
    except Exception as e:
        return f"failed to read memory: {e}"


def report_memory_gpu():
    """Return a formatted GPU memory usage summary for the process."""
    try:
        import subprocess

        import psutil

        pid = psutil.Process().pid

        process = subprocess.run(
            [
                "nvidia-smi",
                "--query-compute-apps=pid,gpu_uuid,used_memory",
                "--format=csv,noheader,nounits",
            ],
            capture_output=True,
        )
        stdout = process.stdout.decode()

        if not stdout:
            return (
                f"GPU Process memory: {'... ':>6}GB / "
                f"GPU Memory used: {'... ':>6}GB / "
                f"GPU Total memory: {'... ':>6}GB "
            )
        else:
            process_info = {
                int(entries[0]): (entries[1], entries[2])
                for entries in (
                    line.split(", ") for line in stdout.strip().split("\n")
                )
            }
            if pid in process_info:
                gpu_uuid, gpu_process_memory = process_info[pid]
                gpu_process_memory = int(gpu_process_memory)
            else:
                return (
                    f"GPU Process memory: {'... ':>6}GB / "
                    f"GPU Memory used: {'... ':>6}GB / "
                    f"GPU Total memory: {'... ':>6}GB "
                )

        process = subprocess.run(
            [
                "nvidia-smi",
                "--query-gpu=memory.used,memory.total",
                "--format=csv,noheader,nounits",
                f"--id={gpu_uuid}",
            ],
            capture_output=True,
        )
        stdout = process.stdout.decode().strip()
        gpu_memory_used, gpu_memory_total = map(int, stdout.split(", "))

        return (
            f"GPU Process memory: {gpu_process_memory / 1e3:>6.2f}GB / "
            f"GPU Memory used: {gpu_memory_used / 1e3:>6.2f}GB / "
            f"GPU Total memory: {gpu_memory_total / 1e3:>6.2f}GB "
        )
    except Exception as e:
        return f"failed to read gpu memory: {e}"


def autocorrect_kwargs(func=None, valid_kwargs=None):
    """A decorator that suggests the right keyword arguments if you get them
    wrong. Useful for functions with many specific options.

    Parameters
    ----------
    func : callable, optional
        The function to decorate.
    valid_kwargs : sequence[str], optional
        The valid keyword arguments for ``func``, if not given these are
        inferred from the function signature.
    """
    if func is None:
        # decorator with options
        return functools.partial(autocorrect_kwargs, valid_kwargs=valid_kwargs)

    if valid_kwargs is None:
        import inspect

        sig = inspect.signature(func)
        params = sig.parameters
        valid_kwargs = set(params.keys())
    else:
        valid_kwargs = set(valid_kwargs)

    @functools.wraps(func)
    def wrapped(*args, **kwargs):
        wrong_opts = {kw for kw in kwargs if kw not in valid_kwargs}
        if wrong_opts:
            import difflib

            right_opts = (
                difflib.get_close_matches(opt, valid_kwargs, n=3)
                for opt in wrong_opts
            )
            msg = "Option(s) {} not valid.\n Did you mean: {}?".format(
                wrong_opts, ", ".join(map(str, right_opts))
            )
            print(msg)
            raise ValueError(msg)

        return func(*args, **kwargs)

    return wrapped



from collections import OrderedDict
from functools import partial

import numpy as np
import pytest
import xarray as xr
from numpy.testing import assert_allclose

from xyzpy.gen.combo_runner import (
    combo_runner,
    combo_runner_to_ds,
    results_to_ds,
)

from . import (
    foo2_array,
    foo2_array_array,
    foo2_array_bool,
    foo2_zarray1_zarray2,
    foo3_float_bool,
    foo3_scalar,
    foo_array_input,
)

# --------------------------------------------------------------------------- #
# COMBO_RUNNER tests                                                          #
# --------------------------------------------------------------------------- #

_test_combos1 = (
    ("a", [1, 2]),
    ("b", [10, 20, 30]),
    ("c", [100, 200, 300, 400]),
)

_test_expect1 = (
    np.array([1, 2]).reshape((2, 1, 1))
    + np.array([10, 20, 30]).reshape((1, 3, 1))
    + np.array([100, 200, 300, 400]).reshape((1, 1, 4))
)


class TestComboRunner:
    @pytest.mark.parametrize("shuffle", [False, True, 2])
    def test_simple(self, shuffle):
        x = combo_runner(foo3_scalar, _test_combos1, shuffle=shuffle)
        assert_allclose(x, _test_expect1)

    def test_progbars(self):
        combo_runner(foo3_scalar, _test_combos1, verbosity=2)

    def test_dict(self):
        combos = OrderedDict(_test_combos1)
        x = combo_runner(foo3_scalar, combos)
        assert_allclose(x, _test_expect1)

    def test_single_combo(self):
        combos = [("a", [1, 2])]
        x = combo_runner(partial(foo3_scalar, b=20, c=300), combos)
        assert_allclose(x, [321, 322])

    def test_single_combo_single_tuple(self):
        combos = ("a", [1, 2])
        constants = {"b": 20, "c": 300}
        x = combo_runner(foo3_scalar, combos, constants=constants)
        assert_allclose(x, [321, 322])

    def test_multires(self):
        combos = _test_combos1
        x, y = combo_runner(foo3_float_bool, combos, split=True)
        xn = _test_expect1
        yn = (
            np.array([1, 2]).reshape((2, 1, 1))
            % np.array([2] * 24).reshape((2, 3, 4))
        ) == 0
        assert_allclose(x, xn)
        assert_allclose(y, yn)

    @pytest.mark.parametrize("shuffle", [False, True, 2])
    @pytest.mark.parametrize("parallel", [False, True])
    @pytest.mark.parametrize("fn", (foo3_scalar,))
    def test_parallel_basic(self, parallel, fn, shuffle):
        x = combo_runner(
            fn,
            _test_combos1,
            num_workers=2,
            parallel=parallel,
            shuffle=shuffle,
        )
        assert_allclose(x, _test_expect1)

    @pytest.mark.parametrize(
        "executor", ["cf-process", "cf-thread", "mp-process", "mp-thread"]
    )
    @pytest.mark.parametrize("fn", (foo3_scalar,))
    def test_executor_basic(self, executor, fn):
        import concurrent.futures as cf
        import multiprocessing as mp

        executor = {
            "cf-process": cf.ProcessPoolExecutor,
            "cf-thread": cf.ThreadPoolExecutor,
            "mp-process": mp.Pool,
            "mp-thread": mp.pool.ThreadPool,
        }[executor](2)
        x = combo_runner(fn, _test_combos1, executor=executor)
        assert_allclose(x, _test_expect1)

    @pytest.mark.parametrize("parallel", [False, True])
    def test_parallel_multires(self, parallel):
        x = combo_runner(
            foo3_float_bool,
            _test_combos1,
            num_workers=2,
            split=True,
            parallel=parallel,
        )
        assert_allclose(x[0], _test_expect1)
        assert np.all(np.asarray(x[1])[1, ...])

    @pytest.mark.parametrize("parallel", [False, True])
    def test_parallel_dict(self, parallel):
        combos = OrderedDict(_test_combos1)
        x = [
            *combo_runner(
                foo3_scalar,
                combos,
                num_workers=2,
                parallel=parallel,
                verbosity=2,
            )
        ]
        assert_allclose(x, _test_expect1)


class TestCombosToDS:
    def test_simple(self):
        results = [1, 2, 3]
        combos = [("a", [1, 2, 3])]
        var_names = ["sum"]
        ds = results_to_ds(
            results, combos, var_names, var_dims={"sum": ()}, var_coords={}
        )
        assert ds["sum"].data.dtype == int


class TestComboRunnerToDS:
    @pytest.mark.parametrize("shuffle", [False, True, 2])
    def test_basic(self, shuffle):
        combos = _test_combos1
        ds = combo_runner_to_ds(
            foo3_scalar, combos, var_names=["bananas"], shuffle=shuffle
        )
        assert ds.sel(a=2, b=30, c=400)["bananas"].data == 432

    def test_multiresult(self):
        ds = combo_runner_to_ds(
            foo3_float_bool, _test_combos1, var_names=["bananas", "cakes"]
        )
        assert ds.bananas.data.dtype == int
        assert ds.cakes.data.dtype == bool
        assert ds.sel(a=2, b=30, c=400)["bananas"].data == 432
        assert ds.sel(a=1, b=10, c=100)["bananas"].data == 111
        assert ds.sel(a=2, b=30, c=400)["cakes"].data
        assert not ds.sel(a=1, b=10, c=100)["cakes"].data

    def test_arrayresult(self):
        combos = (("a", [1, 2]), ("b", [10, 20, 30]))
        ds = combo_runner_to_ds(
            foo2_array,
            combos,
            var_names="bananas",
            var_dims={"bananas": ["sugar"]},
            var_coords={"sugar": [*range(10)]},
        )
        assert ds.bananas.data.dtype == float
        assert_allclose(
            ds.sel(a=2, b=30)["bananas"].data,
            [32.0, 32.1, 32.2, 32.3, 32.4, 32.5, 32.6, 32.7, 32.8, 32.9],
        )

    def test_array_and_single_result(self):
        combos = (("a", [1, 2]), ("b", [10, 20, 30]))
        ds = combo_runner_to_ds(
            foo2_array_bool,
            combos,
            var_names=["bananas", "ripe"],
            var_dims=(["sugar"], []),
            var_coords={"sugar": [*range(10, 20)]},
        )
        assert ds.ripe.data.dtype == bool
        assert ds.sel(a=2, b=30, sugar=14)["bananas"].data == 32.4
        with pytest.raises((KeyError, ValueError)):
            ds["ripe"].sel(sugar=12)

    def test_dict_return(self):
        def f(a, b):
            return {"apples": a + b, "oranges": a - b}

        ds = combo_runner_to_ds(
            f,
            combos=[("a", [1, 2]), ("b", [30, 40])],
            var_names=None,
        )
        assert "apples" in ds
        assert "oranges" in ds
        assert ds.apples.data.dtype == int
        assert ds.oranges.data.dtype == int
        assert_allclose(ds.sel(a=2, b=30).apples.data, 32)
        assert_allclose(ds.sel(a=2, b=30).oranges.data, -28)

    def test_single_string_var_names_with_no_var_dims(self):
        combos = ("a", [1, 2, 3])
        ds = combo_runner_to_ds(
            foo3_scalar, combos, constants={"b": 10, "c": 100}, var_names="sum"
        )
        assert_allclose(ds["sum"].data, np.array([111, 112, 113]))

    def test_double_array_return_with_same_dimensions(self):
        combos = (("a", [1, 2]), ("b", [10, 20, 30]))
        ds = combo_runner_to_ds(
            foo2_array_array,
            combos,
            var_names=["apples", "oranges"],
            var_dims={("apples", "oranges"): ["seeds"]},
            var_coords={"seeds": [*range(5)]},
        )
        assert ds.oranges.data.dtype == int
        assert_allclose(ds.sel(a=2, b=30).apples.data, [30, 32, 34, 36, 38])

        assert_allclose(ds.sel(a=2, b=30).oranges.data, [30, 28, 26, 24, 22])
        assert "seeds" in ds.apples.coords
        assert "seeds" in ds.oranges.coords

    def test_double_array_return_with_no_given_dimensions(self):
        ds = combo_runner_to_ds(
            foo2_array_array,
            combos=[("a", [1, 2]), ("b", [30, 40])],
            var_names=["array1", "array2"],
            var_dims=[["auto"], ["auto"]],
        )
        assert (
            ds["auto"].data.dtype == int or ds["auto"].data.dtype == np.int64
        )
        assert_allclose(ds["auto"].data, [0, 1, 2, 3, 4])

    def test_complex_output(self):
        ds = combo_runner_to_ds(
            foo2_zarray1_zarray2,
            combos=[("a", [1, 2]), ("b", [30, 40])],
            var_names=["array1", "array2"],
            var_dims=[["auto"], ["auto"]],
        )
        assert ds["array1"].data.size == 2 * 2 * 5
        assert ds["array2"].data.size == 2 * 2 * 5
        assert ds["array1"].data.dtype == complex
        assert ds["array2"].data.dtype == complex
        assert_allclose(
            ds["array1"].sel(a=2, b=30).data, 32 + np.arange(5) * 0.1j
        )
        assert_allclose(
            ds["array2"].sel(a=2, b=30).data, 32 - np.arange(5) * 0.1j
        )

    def test_constants_to_attrs(self):
        ds = combo_runner_to_ds(
            foo3_scalar,
            combos=[("a", [1, 2, 3]), ("c", [100, 200, 300])],
            constants={"b": 20},
            var_names="x",
        )
        assert ds.attrs["b"] == 20

    def test_const_array_to_coord(self):
        ds = combo_runner_to_ds(
            foo_array_input,
            combos=[("a", [1, 2, 3])],
            constants={"t": [10, 20, 30]},
            var_names=["x"],
            var_dims=[["t"]],
        )
        assert "t" in ds.dims
        assert "t" not in ds.attrs

    def test_when_results_are_xobjs(self):
        def fn_ds(a, b):
            ds = xr.Dataset(
                coords={
                    "x": [1, 2],
                    "y": ["foo", "bar", "baz"],
                },
                data_vars={
                    "apples": (["x", "y"], np.tile(10 * a + b, (2, 3))),
                    "lemons": ("y", np.tile(10 * a + b, (3,))),
                },
            )
            return ds

        fds = combo_runner_to_ds(
            fn_ds,
            combos={
                "a": [2, 3, 4, 5],
                "b": [6, 7, 8, 9, 10],
            },
            var_names=None,
        )

        assert fds.sel(a=4, b=9, x=2, y="bar")["apples"].values == 49
        assert fds.sel(a=5, b=7, y="bar")["lemons"].values == 57
        assert "output" not in fds
